import pandas as pd
import seaborn
from PyQt6.QtCore import QAbstractListModel, QModelIndex
from PyQt6.QtCore import Qt, QVariant
from PyQt6.QtGui import QColor

__all__ = ['PandasListModel_roles', 'PandasListModel']


class PandasListModel_roles(QAbstractListModel):

    def __init__(self,
                 data: pd.DataFrame,
                 target_column=0,
                 styler=None,
                 parent=None,
                 align=None,
                 style_on_cell=False,
                 *args,
                 **kwargs):

        super().__init__(parent)

        self._data = data
        self.target_column = target_column
        self.len_data = len(data)

        self.detect_flag = Qt.ItemDataRole.UserRole + 1
        self.fraction = Qt.ItemDataRole.UserRole + 2

    def data(self, index: QModelIndex, role: int = ...):
        row = index.row()

        if role == Qt.ItemDataRole.DisplayRole:
            return self._data.iat[row, self.target_column]
        if role == Qt.ItemDataRole.BackgroundRole:
            return "#ff5468"
        if role == self.detect_flag:
            return self._data.iat[row, 0]

        if role == self.fraction:
            return self._data.iat[row, 1]

    def roleNames(self):
        return {Qt.ItemDataRole.UserRole + 1: b'detect_flag',
                Qt.ItemDataRole.UserRole + 2: b'fraction'}

    def rowCount(self, parent=QModelIndex()):
        r = len(self._data)
        return r


class PandasListModel(QAbstractListModel):
    """Custom class to populate a QTableview with a pandas dataframe"""

    _alignment = Qt.AlignmentFlag.AlignLeft
    _styler = None
    _original_data: pd.DataFrame = None
    _style_on_cell = None

    _data = None
    _col_width = None
    _RowCount = 0
    _ColCount = 0
    target_column = 0

    h_index = None
    v_index = None

    palette_list = seaborn.color_palette(None, 20).as_hex()
    palette_dict = dict()

    customRoles = {}

    def __init__(self,
                 data: pd.DataFrame = None,
                 target_column: int | str = 0,
                 styler=None,
                 parent=None,
                 alignment=None,
                 *args,
                 **kwargs):

        QAbstractListModel.__init__(self, parent)

        self.original_data = data
        self._alignment = alignment
        self._styler = styler

        self.target_column = target_column

    @property
    def original_data(self):
        return self._original_data

    @original_data.setter
    def original_data(self, value):
        self._original_data = value if value is not None else pd.DataFrame()
        self._data = self._original_data

        self._RowCount, self._ColCount = self._original_data.shape

        self.v_index = self._original_data.index
        self.h_index = self._original_data.columns
        self._col_width = self._original_data.map(lambda x: len(str(x)), na_action='ignore').max(axis=0).to_list()

    def setNewData(self,
                   data,
                   *args,
                   **kwargs):
        """Sets new internal data variable and refreshes model"""

        self._alignment = kwargs.get('align', Qt.AlignmentFlag.AlignLeft)

        self._styler = kwargs.get('styler', None)

        self._style_on_cell = kwargs.get('style_on_cell', False)

        self.target_column = kwargs.get('target_column', 0)

        self.original_data = data

        self.layoutChanged.emit()

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):

        if index.isValid() and self._data is not None:

            value = self._data.iat[index.row(), self.target_column]

            if self._styler is not None:
                if role == Qt.ItemDataRole.BackgroundRole:
                    return QColor(self._styler(value, role))

                elif role == Qt.ItemDataRole.ForegroundRole:
                    return QColor(self._styler(value, role))

            if role == Qt.ItemDataRole.DisplayRole:
                return str(value)

            if role == Qt.ItemDataRole.EditRole:
                return str(value)

            if role == Qt.ItemDataRole.UserRole:
                return str(value)

            if role == Qt.ItemDataRole.TextAlignmentRole and self._alignment is not None:
                return QVariant(self._alignment)

        return QVariant()

    def setData(self, index, value, role=Qt.ItemDataRole.EditRole):
        if index.isValid():
            if role == Qt.ItemDataRole.DisplayRole or role == Qt.ItemDataRole.EditRole:
                self._data.iat[index.row(), index.column()] = value
                self.dataChanged.emit(index, index)
                return True
        return False

    def rowCount(self, parent=QModelIndex()):
        return self._RowCount

    def columnCount(self, parent=None):
        return self._ColCount

    def headerData(self, index, orientation, role=Qt.ItemDataRole.DisplayRole):
        if (orientation == Qt.Orientation.Horizontal
                and role == Qt.ItemDataRole.DisplayRole):
            return QVariant(str(self.h_index[index]))

        if (orientation == Qt.Orientation.Vertical
                and role == Qt.ItemDataRole.DisplayRole):
            return QVariant(str(self.v_index[index]))

        return QVariant()

    def flags(self, index):
        return Qt.ItemFlag.ItemIsSelectable | Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsDragEnabled

    def getData(self) -> pd.Series:
        """Retrieve data as series"""
        return self._data[self.target_column]



import os
import sys
from functools import partial

from PyQt6 import QtCore, QtWidgets, QtGui

if sys.version_info.major >= 3:
    long = int

active_dcc_is_maya = "maya" in os.path.basename(sys.executable).lower()
active_dcc_is_houdini = "houdini" in os.path.basename(sys.executable).lower()

standalone_app_window = None

resources_folder = os.path.join(os.path.dirname(os.path.dirname(__file__)), "resources")


def get_app_window():
    top_window = standalone_app_window

    if active_dcc_is_maya:
        from maya import OpenMayaUI as omui
        maya_main_window_ptr = omui.MQtUtil().mainWindow()
        top_window = wrapInstance(long(maya_main_window_ptr), QtWidgets.QMainWindow)

    elif active_dcc_is_houdini:
        import hou
        top_window = hou.qt.mainWindow()

    return top_window


class CoreToolWindow(QtWidgets.QMainWindow):
    def __init__(self, parent=None):
        if parent is None:
            parent = get_app_window()
        super(CoreToolWindow, self).__init__(parent)

        self.ui = None
        self.setWindowTitle(self.__class__.__name__)

    def main(self, *args, **kwargs):
        self.show()

    #########################################################
    # convenience functions to make a simple button layout

    def ensure_main_layout(self):
        if self.ui is None:
            main_widget = QtWidgets.QWidget()
            main_layout = QtWidgets.QVBoxLayout()
            main_widget.setLayout(main_layout)
            self.ui = main_widget
            self.setCentralWidget(main_widget)

    def add_button(self, text, command, clicked_args=None):
        self.ensure_main_layout()

        main_layout = self.ui.layout()

        btn = QtWidgets.QPushButton(text)
        btn.setSizePolicy(QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Minimum)
        main_layout.addWidget(btn)

        if clicked_args:
            btn.clicked.connect(partial(command, *clicked_args))
        else:
            btn.clicked.connect(command)


class WindowCache:
    window_instances = {}


if active_dcc_is_maya:

    from maya.app.general.mayaMixin import MayaQWidgetDockableMixin
    from maya import OpenMayaUI as omui
    from maya import cmds


    class ToolWindow(MayaQWidgetDockableMixin, CoreToolWindow):
        def __init__(self, parent=None):
            if parent is None:
                parent = get_app_window()
            super(ToolWindow, self).__init__(parent=parent)
            self.setAttribute(QtCore.Qt.WidgetAttribute.WA_DeleteOnClose, True)

            class_name = self.__class__.__name__
            self.setObjectName(class_name)

        def main(self, restore=False, refresh=False):
            object_name = self.objectName()

            if refresh:
                WindowCache.window_instances.pop(object_name, None)

                workspace_control_name = object_name + "WorkspaceControl"
                if cmds.workspaceControl(workspace_control_name, q=True, exists=True):
                    cmds.workspaceControl(workspace_control_name, e=True, close=True)
                    cmds.deleteUI(workspace_control_name, control=True)

            if restore:
                restored_control = omui.MQtUtil.getCurrentParent()

            launch_ui_script = "import {module}; {module}.{class_name}().main(restore=True)".format(
                module=self.__class__.__module__,
                class_name=self.__class__.__name__
            )

            window_instance = WindowCache.window_instances.get(object_name)
            if not window_instance:
                window_instance = self
                WindowCache.window_instances[object_name] = window_instance

            if restore:
                mixin_ptr = omui.MQtUtil.findControl(window_instance.objectName())
                omui.MQtUtil.addWidgetToMayaLayout(long(mixin_ptr), long(restored_control))
            else:
                window_instance.show(dockable=True, height=600, width=480, uiScript=launch_ui_script)

            return window_instance

else:
    ToolWindow = CoreToolWindow


def build_menu_from_action_list(actions, menu=None, is_sub_menu=False):
    if not menu:
        menu = QtWidgets.QMenu()

    for action in actions:
        if action == "-":
            menu.addSeparator()
            continue

        for action_title, action_command in action.items():
            if action_title == "RADIO_SETTING":
                # Create RadioButtons for QSettings object
                settings_obj = action_command.get("settings")  # type: QtCore.QSettings
                settings_key = action_command.get("settings_key")  # type: str
                choices = action_command.get("choices")  # type: list
                default_choice = action_command.get("default")  # type: str
                on_trigger_command = action_command.get("on_trigger_command")  # function to trigger after setting value

                # Has choice been defined in settings?
                item_to_check = settings_obj.value(settings_key)

                # If not, read from default option argument
                if not item_to_check:
                    item_to_check = default_choice

                grp = QtGui.QActionGroup(menu)
                for choice_key in choices:
                    action = QtGui.QAction(choice_key, menu)
                    action.setCheckable(True)

                    if choice_key == item_to_check:
                        action.setChecked(True)

                    action.triggered.connect(partial(
                        set_settings_value,
                        settings_obj,
                        settings_key,
                        choice_key,
                        on_trigger_command
                    ))
                    menu.addAction(action)
                    grp.addAction(action)

                grp.setExclusive(True)
                continue

            if isinstance(action_command, list):
                sub_menu = menu.addMenu(action_title)
                build_menu_from_action_list(action_command, menu=sub_menu, is_sub_menu=True)
                continue

            atn = menu.addAction(action_title)
            atn.triggered.connect(action_command)

    if not is_sub_menu:
        cursor = QtGui.QCursor()
        menu.exec(cursor.pos())

    return menu


def set_settings_value(settings_obj, key, value, post_set_command=None):
    settings_obj.setValue(key, value)
    if post_set_command:
        post_set_command()


class QtPathWidget(QtWidgets.QWidget):
    path_changed = QtCore.pyqtSignal(str)

    def __init__(
            self,
            parent=None,
            settings_name="_DefaultQtPathWidgetSettings",
            start_dir="",
            file_filter="",
            use_directory_dialog=False,
            relative_to_path="",
            recent_paths_amount=30,
            only_show_existing_recent_paths=False
    ):
        """
        QWidget for file paths.

        Includes "browse" button and list of recent file paths.

        :param parent: Qt Parent
        :param settings_name: name for settings .ini
        :param start_dir: start folder for file dialog
        :param file_filter: filter for the QFileDialog
        :param use_directory_dialog: browse for folder instead of file path
        :param relative_to_path: show paths relative to this path
        :param recent_paths_amount: clamp recent paths to this amount
        :param only_show_existing_recent_paths:
        """

        super(QtPathWidget, self).__init__(parent)

        # safety convert, just in case we get passed negative float values for some reason
        recent_paths_amount = int(abs(recent_paths_amount))

        self.start_dir = start_dir
        self.relative_to_path = relative_to_path
        self.relative_to_path_drive = os.path.splitdrive(self.relative_to_path)[0]
        self.use_directory_dialog = use_directory_dialog
        self.recent_paths_amount = recent_paths_amount

        # settings object to store data between sessions
        self._settings = QtPathWidgetSettings(
            identifier=settings_name,
            recent_paths_amount=recent_paths_amount,
            relative_to_path=relative_to_path,
        )

        main_layout = QtWidgets.QHBoxLayout()
        main_layout.setContentsMargins(0, 0, 0, 0)

        # surprise, it's a QComboBox for the path display
        self.path_CB = QtWidgets.QComboBox()
        self.path_CB.addItems(self._settings.get_recent_paths(only_existing=only_show_existing_recent_paths))
        self.path_CB.setEditable(True)
        self.path_CB.setCurrentText("")
        self.path_CB.setSizePolicy(QtWidgets.QSizePolicy.Policy.Ignored, QtWidgets.QSizePolicy.Policy.Fixed)
        self.path_CB.currentTextChanged.connect(self.path_changed)
        main_layout.addWidget(self.path_CB)

        # Browse path button
        self.browse_file_BTN = QtWidgets.QPushButton("...")
        self.browse_file_BTN.setMaximumWidth(40)
        self.browse_file_BTN.setSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
        self.browse_file_BTN.clicked.connect(self.open_dialog_and_set_path)
        main_layout.addWidget(self.browse_file_BTN)

        self.setLayout(main_layout)

        # create path dialog instance for later use
        self.path_dialog = QtWidgets.QFileDialog()

        if self.use_directory_dialog:
            self.path_dialog.setFileMode(QtWidgets.QFileDialog.FileMode.Directory)

        self.path_dialog.setNameFilter(file_filter)

    def open_dialog_and_set_path(self, emit_change_signal=True):
        file_path = self.get_dialog_path()
        if file_path:
            self.set_path(file_path, emit_change_signal)
            return file_path

    def get_dialog_path(self):
        # set starting directory for dialog
        start_dir = self.start_dir

        # no start_dir specified, use folder of most recent path
        if not start_dir:
            start_dir = str(self._settings.value(QtPathWidgetSettings.key_most_recent_dir, defaultValue=""))

        self.path_dialog.setDirectory(start_dir)

        # display dialog
        if self.path_dialog.exec_():
            selected_paths = self.path_dialog.selectedFiles()
            self._settings.setValue(QtPathWidgetSettings.key_most_recent_dir, os.path.dirname(selected_paths[0]))
            return selected_paths[0]

    def set_path(self, path, emit_change_signal=True):
        """
        Set path display in ComboBox and store in settings

        :param path:
        :return:
        """
        self._settings.add_recent_path(path)  # store full path in settings, then convert to relative if desired

        # convert to relative path
        path_drive = os.path.splitdrive(path)[0]
        if self.relative_to_path and path_drive == self.relative_to_path_drive:
            path = os.path.relpath(path, self.relative_to_path)

        # if path has already been added to ComboBox, remove the old one
        path_index_map = {self.path_CB.itemText(i): i for i in range(self.path_CB.count())}
        if path in path_index_map.keys():
            self.path_CB.removeItem(path_index_map[path])

        # add to ComboBox and set as active path
        self.path_CB.insertItem(0, path)
        self.path_CB.setCurrentIndex(0)

        # clamp amount of recent paths
        while self.path_CB.count() > self.recent_paths_amount:
            self.path_CB.removeItem(self.recent_paths_amount)

        if emit_change_signal:
            self.path_changed.emit(path)

    def path(self):
        """get path from widget"""
        current_path = self.path_CB.currentText()

        # join with relative_to_path if it's a relative path
        if self.relative_to_path and os.path.splitdrive(current_path)[0] != "":
            return os.path.abspath(os.path.join(self.relative_to_path, current_path))

        return current_path

    ###################################################
    # Convenience functions for replacing LineEdit with this widget

    def text(self):
        return self.path()

    def setText(self, value):
        self.set_path(value)


class QtPathWidgetSettings(QtCore.QSettings):
    key_recent_paths = "recent_paths"
    key_most_recent_dir = "most_recent_dir"

    def __init__(self, identifier="_DefaultQtPathWidgetSettings", recent_paths_amount=30, relative_to_path=""):

        # %appdata%\QtPathWidget\_DefaultQtPathWidgetSettings.ini

        super(QtPathWidgetSettings, self).__init__(
            QtCore.QSettings.Format.IniFormat,
            QtCore.QSettings.Scope.UserScope,
            'QtPathWidget',
            identifier
        )
        self.recent_paths_amount = recent_paths_amount
        self.relative_to_path = relative_to_path
        self.relative_to_path_drive = os.path.splitdrive(self.relative_to_path)[0]

    def get_recent_paths(self, full_paths=False, only_existing=False):
        """
        Get recent paths from settings

        :param full_paths: skip converting to relative paths
        :param only_existing: only return full existing paths
        :return:
        """
        paths = self.value(self.key_recent_paths)
        if not isinstance(paths, list):
            if paths:
                # QSettings ini sometimes has trouble reading data types
                paths = str(paths).split(", ")
            else:
                paths = []

        if only_existing:
            paths = [p for p in paths if os.path.exists(p)]

        # convert to relative paths before returning
        if not full_paths and self.relative_to_path:
            relative_paths = []

            for full_path in paths:
                if os.path.splitdrive(full_path)[0] != self.relative_to_path_drive:
                    # if the path is on a separate drive then we can't get a relative path
                    relative_paths.append(full_path)
                else:
                    relative_paths.append(os.path.relpath(full_path, self.relative_to_path))

            paths = relative_paths

        return paths

    def add_recent_path(self, path):
        """
        Add path to recent paths in settings

        :param path: <str>
        :return:
        """
        recent_paths = self.get_recent_paths(full_paths=True)

        # remove path from recent_paths if it's been added previously
        if path in recent_paths:
            recent_paths.remove(path)

        recent_paths.insert(0, path)

        if len(recent_paths) > self.recent_paths_amount:  # clamp amount of paths
            recent_paths = recent_paths[:self.recent_paths_amount]

        self.setValue(self.key_recent_paths, recent_paths)



import re
import typing

from PyQt6 import QtCore
from PyQt6.QtCore import QModelIndex, Qt
from PyQt6.QtGui import QColor
from matplotlib import colors


class QJsonTreeItem(object):
    def __init__(self, parent=None):
        self._parent = parent

        self._key = ""
        self._value = ""
        self._type = None
        self._children = list()

    def appendChild(self, item):
        self._children.append(item)

    def child(self, row):
        return self._children[row]

    def parent(self):
        return self._parent

    def childCount(self):
        return len(self._children)

    def row(self):
        return (
            self._parent._children.index(self)
            if self._parent else 0
        )

    @property
    def key(self):
        return self._key

    @key.setter
    def key(self, key):
        self._key = key

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, typ):
        self._type = typ

    @classmethod
    def load(self, value, parent=None, sort=True):

        rootItem = QJsonTreeItem(parent)
        rootItem.key = "root"

        if isinstance(value, dict):
            items = (
                sorted(value.items())
                if sort else value.items()
            )

            for key, value in items:
                child = self.load(value, rootItem)
                child.key = key
                child.type = type(value)
                rootItem.appendChild(child)

        elif isinstance(value, list):
            for index, value in enumerate(value):
                child = self.load(value, rootItem)
                child.key = index
                child.type = type(value)
                rootItem.appendChild(child)

        else:
            rootItem.value = value
            rootItem.type = type(value)

        return rootItem


class QJsonModel(QtCore.QAbstractItemModel):

    def __init__(self, parent=None, edit_columns=None):
        super(QJsonModel, self).__init__(parent)

        self.editable_cols = edit_columns

        self._rootItem = QJsonTreeItem()
        self._headers = ("key", "value")

    def clear(self):
        self.load({})

    def load(self, document):
        """Load from dictionary

        Arguments:
            document (dict): JSON-compatible dictionary

        """

        assert isinstance(document, (dict, list, tuple)), (
                "`document` must be of dict, list or tuple, "
                "not %s" % type(document)
        )

        self._data = document
        self.beginResetModel()

        self._rootItem = QJsonTreeItem.load(document)
        self._rootItem.type = type(document)

        self.endResetModel()

        return True

    def json(self, root=None):
        """Serialise model as JSON-compliant dictionary

        Arguments:
            root (QJsonTreeItem, optional): Serialise from here
                defaults to the the top-level item

        Returns:
            model as dict

        """
        root = root or self._rootItem
        return self.genJson(root)

    def getLevel(self, parent_item):
        complete = False
        i = 0

        while not complete:
            if parent_item == None:
                return i
            else:
                i += 1

    def data(self, index: QModelIndex, role: int = ...) -> typing.Any:
        if not index.isValid():
            return None

        item = index.internalPointer()

        if role == QtCore.Qt.ItemDataRole.DisplayRole:
            if index.column() == 0:
                return item.key

            if index.column() == 1:
                return item.value

        if role == QtCore.Qt.ItemDataRole.EditRole:
            if index.column() == 1:
                return item.value

        if role == Qt.ItemDataRole.BackgroundRole and index.column() == 1:
            if re.search(r'^#(?:[0-9a-fA-F]{3}){1,2}$', str(item.value)):
                return QColor(item.value)

        if role == Qt.ItemDataRole.ForegroundRole and index.column() == 1:
            if re.search(r'^#(?:[0-9a-fA-F]{3}){1,2}$', str(item.value)):
                return QColor(self.ContrastColor(item.value))

    def setData(self, index: QModelIndex, value: typing.Any, role: int = ...) -> bool:
        if role == QtCore.Qt.ItemDataRole.EditRole:
            if index.column() == 1:
                item = index.internalPointer()
                item.value = str(value)

                self.dataChanged.emit(index, index, [QtCore.Qt.ItemDataRole.EditRole])

                return True

        return False

    def headerData(self, section: int, orientation: Qt.Orientation, role: int = ...) -> typing.Any:
        if role != Qt.ItemDataRole.DisplayRole:
            return None

        elif orientation == Qt.Orientation.Horizontal:
            return self._headers[section]

        else:
            return

    def index(self, row, column, parent=QtCore.QModelIndex()):

        if not self.hasIndex(row, column, parent):
            return QtCore.QModelIndex()

        if not parent.isValid():
            parentItem = self._rootItem
        else:
            parentItem = parent.internalPointer()

        childItem = parentItem.child(row)
        if childItem:

            return self.createIndex(row, column, childItem)
        else:
            return QtCore.QModelIndex()

    def parent(self, index):
        if not index.isValid():
            return QtCore.QModelIndex()

        childItem = index.internalPointer()
        parentItem = childItem.parent()

        if parentItem == self._rootItem:
            return QtCore.QModelIndex()

        return self.createIndex(parentItem.row(), 0, parentItem)

    def rowCount(self, parent=QtCore.QModelIndex()):
        if parent.column() > 0:
            return 0

        if not parent.isValid():
            parentItem = self._rootItem
        else:
            parentItem = parent.internalPointer()

        return parentItem.childCount()

    def columnCount(self, parent=QtCore.QModelIndex()):
        return 2

    def flags(self, index):
        flags = super(QJsonModel, self).flags(index)

        if index.column() in self.editable_cols:
            return Qt.ItemFlag.ItemIsEditable | flags
        else:
            return flags

    def genJson(self, item):
        nchild = item.childCount()

        if item.type is dict:
            document = {}
            for i in range(nchild):
                ch = item.child(i)
                document[ch.key] = self.genJson(ch)
            return document

        elif item.type == list:
            document = []
            for i in range(nchild):
                ch = item.child(i)
                document.append(self.genJson(ch))
            return document

        else:
            return item.value

    def ContrastColor(self, hexColor):
        R, G, B = colors.hex2color(hexColor)

        luminance = (0.299 * R + 0.587 * G + 0.114 * B)

        if (luminance > 0.5):
            d = 0  # // bright colors - black font
        else:
            d = 1  # // dark colors - white font

        return colors.rgb2hex((d, d, d))


import sys

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import (QApplication, QComboBox, QWidget, QVBoxLayout, QStyledItemDelegate, QListView)


class CustomDelegate(QStyledItemDelegate):
    def paint(self, painter, option, index):
        if index.isValid():
            painter.save()
            painter.drawText(option.rect, Qt.AlignmentFlag.AlignLeft, f"Name: {index.data(Qt.ItemDataRole.DisplayRole)}, Value: {index.data(Qt.ItemDataRole.UserRole)}")
            painter.restore()

class CustomView(QListView):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setItemDelegate(CustomDelegate(self))

class MainWindow(QWidget):
    def __init__(self):
        super().__init__()

        self.combo = QComboBox(self)
        self.combo.setView(CustomView(self.combo))

        layout = QVBoxLayout(self)
        layout.addWidget(self.combo)

        self.add_items()

    def add_items(self):
        data = [("Item 1", 10), ("Item 2", 20), ("Item 3", 30)]
        for name, value in data:
            self.combo.addItem(name, value)

if __name__ == "__main__":
    app = QApplication(sys.argv)
    window = MainWindow()
    window.show()
    sys.exit(app.exec())


from PyQt6.QtCore import QEvent, Qt
from PyQt6.QtGui import QPalette, QFontMetrics, QStandardItem
from PyQt6.QtWidgets import QComboBox, QStyledItemDelegate


class CheckableComboBox(QComboBox):
    # Subclass Delegate to increase item height
    class Delegate(QStyledItemDelegate):
        def sizeHint(self, option, index):
            size = super().sizeHint(option, index)
            size.setHeight(20)
            return size

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Make the combo editable to set a custom text, but readonly
        self.setEditable(True)
        self.lineEdit().setReadOnly(True)
        # Make the lineedit the same color as QPushButton
        palette = qApp.palette()
        palette.setBrush(QPalette.ColorRole.Base, palette.button())
        self.lineEdit().setPalette(palette)

        # Use custom delegate
        self.setItemDelegate(CheckableComboBox.Delegate())

        # Update the text when an item is toggled
        self.model().dataChanged.connect(self.updateText)

        # Hide and show popup when clicking the line edit
        self.lineEdit().installEventFilter(self)
        self.closeOnLineEditClick = False

        # Prevent popup from closing when clicking on an item
        self.view().viewport().installEventFilter(self)

    def resizeEvent(self, event):
        # Recompute text to elide as needed
        self.updateText()
        super().resizeEvent(event)

    def eventFilter(self, object, event):

        if object == self.lineEdit():
            if event.type() == QEvent.Type.MouseButtonRelease:
                if self.closeOnLineEditClick:
                    self.hidePopup()
                else:
                    self.showPopup()
                return True
            return False

        if object == self.view().viewport():
            if event.type() == QEvent.Type.MouseButtonRelease:
                index = self.view().indexAt(event.pos())
                item = self.model().item(index.row())

                if item.checkState() == Qt.CheckState.Checked:
                    item.setCheckState(Qt.CheckState.Unchecked)
                else:
                    item.setCheckState(Qt.CheckState.Checked)
                return True
        return False

    def showPopup(self):
        super().showPopup()
        # When the popup is displayed, a click on the lineedit should close it
        self.closeOnLineEditClick = True

    def hidePopup(self):
        super().hidePopup()
        # Used to prevent immediate reopening when clicking on the lineEdit
        self.startTimer(100)
        # Refresh the display text when closing
        self.updateText()

    def timerEvent(self, event):
        # After timeout, kill timer, and reenable click on line edit
        self.killTimer(event.timerId())
        self.closeOnLineEditClick = False

    def updateText(self):
        texts = []
        for i in range(self.model().rowCount()):
            if self.model().item(i).checkState() == Qt.CheckState.Checked:
                texts.append(self.model().item(i).text())
        text = ", ".join(texts)

        # Compute elided text (with "...")
        metrics = QFontMetrics(self.lineEdit().font())
        elidedText = metrics.elidedText(text, Qt.TextElideMode.ElideRight, self.lineEdit().width())
        self.lineEdit().setText(elidedText)

    def addItem(self, text, data=None):
        item = QStandardItem()
        item.setText(text)
        if data is None:
            item.setData(text)
        else:
            item.setData(data)
        item.setFlags(Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsUserCheckable)
        item.setData(Qt.CheckState.Unchecked, Qt.ItemDataRole.CheckStateRole)
        self.model().appendRow(item)

    def addItems(self, texts, datalist=None):
        for i, text in enumerate(texts):
            try:
                data = datalist[i]
            except (TypeError, IndexError):
                data = None
            self.addItem(text, data)

    def currentData(self):
        # Return the list of selected items data
        res = []
        for i in range(self.model().rowCount()):
            if self.model().item(i).checkState() == Qt.CheckState.Checked:
                res.append(self.model().item(i).data())
        return res



import logging
import pathlib

from collections.abc import Callable
import pandas as pd
from PyQt6.QtCore import pyqtSignal, QThread, QTimer, QObject
from PyQt6.QtWidgets import QFileDialog, QApplication, QToolButton

from power_qt.buttons.active_button import ActiveButton
from power_qt.other.default_fields import DefaultFieldsWidget
from abc import ABC, abstractmethod

logger = logging.getLogger('__main__')
App = QApplication([])


class masterBaseClass(ABC):
    # import guicanvas & guitable

    _module_sources = []

    @abstractmethod
    def ConfigChanged(self, newname) -> None:
        """Target function for performing refresh of widget data when base data changes"""
        return any([name in self._module_sources for name in newname])

    @abstractmethod
    def start(self) -> None:
        """Generic startup function, perform all data modifications for widget here only"""
        pass

    def _set_default_dictionary(self, new_obj: dict):
        """Generic function for setting a class' default fields"""
        pass

    def create_editor(self, path, addable_dictionary: dict, window_name='Properties'):
        s_path = pathlib.Path(path)
        parent1 = s_path.parents[1].name
        parent2 = s_path.parents[0].name

        t_path = Core.cache / parent1 / parent2 / f'{s_path.stem}.json'

        t_path.parent.mkdir(parents=True, exist_ok=True)

        self.editor = DefaultFieldsWidget(self)
        self.editor.setWindowTitle(window_name)
        self.editor._path = t_path
        self.editor.add_pages(addable_dictionary)

    def edit_styles(self):
        self.editor.show()


class HelperBase():
    _return_signal = pyqtSignal(tuple())
    _timer = QTimer()
    _saveDialog = QFileDialog()
    __thread = None

    def addToolButton(self):
        self.toolb = QToolButton()
        self.layout().addWidget(self.toolb)

    def save_file(self,
                  target: Callable[[any], any],
                  target_kwargs: dict = None,
                  dialog_kwargs: dict = None,
                  *args,
                  **kwargs):
        """
                Helper function for moving a save process into a side thread while saving.

                For filename destinations only.

                target == callable(savepath, *args, **kwargs)

                returns to _thread_run_return
                """
        if dialog_kwargs is None:
            dialog_kwargs = {'filter': 'Excel File (*.xlsx)'}

        sender = self.sender()

        savedialog = self._saveDialog.getSaveFileName

        self.__save_helper(sender, target, savedialog, target_kwargs, dialog_kwargs, *args, **kwargs)

    def save_folder(self,
                    target: Callable[[any], any],
                    target_kwargs: dict = None,
                    dialog_kwargs: dict = None,
                    *args,
                    **kwargs):
        """
                Helper function for moving a save process into a side thread while saving.

                For folderpath destinations only.

                target == callable(folderpath, *args, **kwargs)

                returns to _thread_run_return
                """

        sender = self.sender()
        savedialog = self._saveDialog.getExistingDirectory

        self.__save_helper(sender, target, savedialog, target_kwargs, dialog_kwargs, *args, **kwargs)

    def __save_helper(self,
                      sender: QObject,
                      target: Callable[[any], any],
                      savedialog,
                      target_kwargs: dict = None,
                      dialog_kwargs: dict = None,
                      *args,
                      **kwargs):

        """Private method for assembling thread and savedialog from save_file and save_folder"""

        if dialog_kwargs is None:
            dialog_kwargs = dict()

        if target_kwargs is None:
            target_kwargs = dict()

        def_dialog_kwargs = {'directory': '',
                             'caption': 'Save File'}

        dialog_kwargs = def_dialog_kwargs | dialog_kwargs

        savepath = savedialog(parent=None, **dialog_kwargs)

        savepath = savepath[0] if isinstance(savepath, tuple) else savepath

        if savepath == '' and isinstance(sender, ActiveButton):
            sender.cancelled()
            return

        self.__create_saving_thread(sender, target, (savepath,), target_kwargs)

    def __create_saving_thread(self, sender_obj, target, savepath, target_kwargs):
        self.myThread = MasterThread(target, target_args=savepath, target_kwargs=target_kwargs)

        self.myThread.finished.connect(lambda returnable: self._thread_run_return(returnable))
        self.myThread.errored.connect(lambda returnable: self._thread_run_return(returnable))

        if isinstance(sender_obj, ActiveButton):
            self.myThread.started.connect(sender_obj.running)
            self.myThread.errored.connect(sender_obj.errored)
            self.myThread.finished.connect(sender_obj.finished)

        self.myThread.start()

    def threaded_run(self,
                     target=None,
                     target_args: tuple | list = None,
                     target_kwargs: dict = None):

        button = self.sender()

        if target_args is None:
            target_args = tuple()

        if target_kwargs is None:
            target_kwargs = dict()

        self.__thread = MasterThread(target=target,
                                     target_args=target_args,
                                     target_kwargs=target_kwargs)

        self.__thread.finished.connect(self._thread_run_return)
        self.__thread.errored.connect(self._thread_error_return)

        if isinstance(button, ActiveButton):
            self.__thread.started.connect(button.running)
            self.__thread.errored.connect(button.errored)
            self.__thread.finished.connect(button.finished)

        self.__thread.start()

    def _thread_run_return(self, *args, **kwargs):
        pass

    def _thread_error_return(self, *args, **kwargs):
        logger.exception(args[0])


class MasterThread(QThread):
    finished = pyqtSignal(tuple)
    errored = pyqtSignal(Exception)
    progress = pyqtSignal(int)
    returnable = None

    def __init__(self, target=None, target_args=None, target_kwargs=None):
        super().__init__()
        self.target = target
        self.args = () if target_args is None else target_args
        self.kwargs = {} if target_kwargs is None else target_kwargs
        self.kwargs['progress'] = self.progress

    def run(self):
        try:

            returnable = self.target(*self.args, **self.kwargs)

            if not isinstance(returnable, tuple):
                returnable = (returnable,)

            self.finished.emit(returnable)

        except Exception as E:
            logger.exception(E)
            self.errored.emit(E)



from PyQt6.QtCore import Qt, pyqtSignal
from PyQt6.QtWidgets import QListWidget, QAbstractItemView, QWidget, QLabel, QListWidgetItem, \
    QDoubleSpinBox, QHBoxLayout


class DragDropListWidget(QListWidget):
    currentCountChanged = pyqtSignal()
    trigResort = pyqtSignal()

    """Generic List Widget for holding Movable Items"""

    def __init__(self, parent=None):
        super(QListWidget, self).__init__(parent)
        # self.setDragDropMode(QAbstractItemView.DragDrop)
        self.setDefaultDropAction(Qt.DropAction.MoveAction)
        self.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        # self.setAcceptDrops(True)
        self.model().rowsInserted.connect(self.handleRowsInserted, Qt.ConnectionType.QueuedConnection)
        self.model().rowsRemoved.connect(self.handleRowsInserted)

        self.maxItems = 9999

    def baseItemWidget(self, widget):
        self._item_widget = widget

    def addItems(self, items: list[tuple]):
        for i, data in enumerate(sorted(items)):
            myQListWidgetItem = QListWidgetItem(self)
            # store the data needed to create/re-create the custom widget
            myQListWidgetItem.setData(Qt.ItemDataRole.UserRole, (i, data))
            self.addItem(myQListWidgetItem)

    def handleRowsInserted(self, parent, first, last):
        for index in range(first, last + 1):
            item = self.item(index)
            widget = QCustomSpin()

            widget.setNew(item)
            item.setSizeHint(widget.sizeHint())
            self.setItemWidget(item, widget)
            item.setText('')

    def dropEvent(self, event):
        if self.count() == self.maxItems and event.source() != self:
            event.setDropAction(Qt.DropAction.IgnoreAction)
        else:
            super().dropEvent(event)

    def getItems(self, role: int | None = None):
        return [self.itemWidget(self.item(x)).data(role) for x in range(self.count())]


class GenericItemWidget(QWidget):
    def __init__(self, parent=None):
        super(GenericItemWidget, self).__init__(parent)
        self.user_role = None
        self.display_role = None
        self.edit_role = None

        self.FullLayout = QHBoxLayout()
        self.setLayout(self.FullLayout)

    def setNew(self, item: QListWidgetItem, *args):
        self.display_role = item.data(Qt.ItemDataRole.DisplayRole)

        self.edit_role = item.data(Qt.ItemDataRole.EditRole)

        self.user_role = item.data(Qt.ItemDataRole.UserRole)

        label = QLabel(self.display_role)
        self.FullLayout.addWidget(label)

    def data(self, role, *args, **kwargs):
        return {Qt.ItemDataRole.DisplayRole: self.display_role,
                Qt.ItemDataRole.EditRole: self.edit_role,
                Qt.ItemDataRole.UserRole: self.user_role}

class QCustomSpin(GenericItemWidget):
    """Custom List widget item class"""

    def __init__(self, parent=None):
        super(QCustomSpin, self).__init__(parent)

    def setNew(self, item: QListWidgetItem, *args):
        super().setNew(item)

        self.spin = QDoubleSpinBox()
        self.spin.setSingleStep(0.01)
        self.spin.setValue(1)
        self.layout().addWidget(self.spin)

    def data(self, role, *args, **kwargs):
        return {self.display_role: self.spin.value()}



from PyQt6 import QtWidgets, QtGui
from PyQt6.QtGui import QIcon, QPixmap
from PyQt6.QtWidgets import QTabBar, QTabWidget, QMenu

from power_qt.other.default_fields import DefaultFieldsWidget

__all__ = ['StylizedTabWidget']


class _TabBar(QTabBar):
    colorCodes = {}

    def paintEvent(self, event):
        style = self.style()
        painter = QtGui.QPainter(self)
        option = QtWidgets.QStyleOptionTab()
        for index in range(self.count()):
            self.initStyleOption(option, index)

            bgcolor = QtGui.QColor(self.colorCodes.get(self.tabText(index), '#FFFFFF'))

            option.palette.setColor(QtGui.QPalette.ColorRole.Window, bgcolor)
            option.palette.setColor(QtGui.QPalette.ColorRole.Button, bgcolor)
            style.drawControl(QtWidgets.QStyle.ControlElement.CE_TabBarTab, option, painter)


class StylizedTabWidget(QTabWidget):
    icons = {}

    def __init__(self, parent=None):
        super(StylizedTabWidget, self).__init__(parent)
        self.setTabBar(_TabBar(self))

        self.context_menu = QMenu(self)
        action1 = self.context_menu.addAction('Customize Tabs')
        action1.triggered.connect(self.showEditor)

        attributes = {'Icons': StylizedTabWidget.icons,
                      'Colors': _TabBar.colorCodes}

        self.editor = DefaultFieldsWidget(self)
        self.editor.setWindowTitle('Name')
        self.editor.add_pages(attributes)

    def _set_default_dictionary(self, new_obj):
        StylizedTabWidget.icons.update(new_obj.get('Icons', {}))
        _TabBar.colorCodes.update(new_obj.get('Colors', {}))

    def showEditor(self):
        self.editor.show()

    def contextMenuEvent(self, a0):
        self.context_menu.exec(a0.globalPos())

    def addTab(self, widget, text):
        QTabWidget.addTab(self, widget, text)

    def setTabText(self, index, text):
        QTabWidget.setTabText(self, index, text)

        tab_icon = self.icons.get(text)

        if tab_icon is not None:
            self.setTabIcon(index, QIcon(QPixmap(str(tab_icon))))

    def refresh_tabs(self):
        for i in range(self.count()):
            pass

    def modIcons(self, tabText=None, icon_path=None, ):
        _TabBar.colorCodes = _TabBar.colorCodes | {tabText: icon_path}

    def modColors(self, tabText=None, color=None):
        _TabBar.colorCodes = _TabBar.colorCodes | {tabText: color}



from PyQt6.QtCore import QPointF
from PyQt6.QtCore import Qt
from PyQt6.QtGui import QTransform, QPainter
from PyQt6.QtWidgets import QTableWidget, QApplication, QTableWidgetItem, QVBoxLayout
from PyQt6.QtWidgets import QWidget
from superqt import QSearchableListWidget, QDoubleRangeSlider, QSearchableComboBox

class SearchableCombo(QSearchableComboBox):
    def __init__(self, parent=None):
        QSearchableComboBox.__init__(self, parent)  # Inherit from QWidget


class SearchableList(QSearchableListWidget):
    def __init__(self, parent=None):
        QSearchableListWidget.__init__(self, parent)  # Inherit from QWidget


class Rangeslider(QDoubleRangeSlider):
    def __init__(self, parent=None):
        QDoubleRangeSlider.__init__(self, parent)  # Inherit from QWidget


class PasteableTableWidget(QTableWidget):
    def __init__(self, parent=None):
        super().__init__(parent)

    def keyPressEvent(self, event):
        if event.key() == Qt.Key.Key_V and (event.modifiers() & Qt.KeyboardModifier.ControlModifier):
            selection = self.selectedIndexes()

            if selection:
                row_anchor = selection[0].row()
                column_anchor = selection[0].column()

                clipboard = QApplication.clipboard()

                rows = clipboard.text().split('\n')
                for indx_row, row in enumerate(rows):
                    values = row.split('\t')
                    for indx_col, value in enumerate(values):
                        item = QTableWidgetItem(value)
                        self.setItem(row_anchor + indx_row, column_anchor + indx_col, item)
            super().keyPressEvent(event)


class DynamicWidget(QWidget):
    """Widget for dynamic GUI widget updating"""

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)
        self.layout = QVBoxLayout()
        self.setLayout(self.layout)

    def update_display_widget(self, widget):
        if self.layout.itemAt(0) is not None:
            self.layout.itemAt(0).widget().setParent(None)
        self.layout.addWidget(widget)

class ImageWidget(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self._image = None

    def setImage(self, image):
        self._image = image
        self.update()

    def paintEvent(self, event):
        if self._image is None or self._image.isNull():
            return
        painter = QPainter(self)
        width = self.width()
        height = self.height()
        imageWidth = self._image.width()
        imageHeight = self._image.height()
        r1 = width / imageWidth
        r2 = height / imageHeight
        r = min(r1, r2)
        x = (width - imageWidth * r) / 2
        y = (height - imageHeight * r) / 2
        painter.setTransform(QTransform().translate(x, y).scale(r, r))
        painter.drawImage(QPointF(0, 0), self._image)



from functools import partial

from .ui_utils import QtWidgets


class BatchNameWidget(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super(BatchNameWidget, self).__init__(*args, **kwargs)

        self._search_replace_widgets = []

        self.main_layout = QtWidgets.QVBoxLayout()
        self.main_layout.setContentsMargins(0, 0, 0, 0)

        self.prefix_line_edit = QtWidgets.QLineEdit()
        self.prefix_line_edit.setPlaceholderText("Prefix")
        self.prefix_line_edit.setClearButtonEnabled(True)
        self.suffix_line_edit = QtWidgets.QLineEdit()
        self.suffix_line_edit.setPlaceholderText("Suffix")
        self.suffix_line_edit.setClearButtonEnabled(True)

        self.add_search_replace_button = QtWidgets.QPushButton("+")
        self.add_search_replace_button.clicked.connect(self.add_search_replace_line)

        self.search_replace_layout = QtWidgets.QVBoxLayout()
        self.search_replace_layout.setContentsMargins(0, 0, 0, 0)

        default_items_layout = QtWidgets.QHBoxLayout()
        default_items_layout.setContentsMargins(0, 0, 0, 0)
        default_items_layout.addWidget(self.prefix_line_edit)
        default_items_layout.addWidget(self.suffix_line_edit)
        default_items_layout.addWidget(self.add_search_replace_button)

        self.main_layout.addLayout(self.search_replace_layout)
        self.main_layout.addLayout(default_items_layout)

        self.setLayout(self.main_layout)

    def add_search_replace_line(self):
        widget = SearchReplaceWidget()
        widget.remove_button.clicked.connect(partial(self.remove_sr_widget, widget))
        self.search_replace_layout.addWidget(widget)
        self._search_replace_widgets.append(widget)

    def remove_sr_widget(self, widget):
        self.search_replace_layout.removeWidget(widget)
        self._search_replace_widgets.remove(widget)
        widget.deleteLater()

    def modify_string(self, input_string):
        output_string = input_string
        for sr_widget in self._search_replace_widgets:  # type: SearchReplaceWidget
            output_string = output_string.replace(
                sr_widget.search_line.text(),
                sr_widget.replace_line.text()
            )
        return "{}{}{}".format(self.prefix_line_edit.text(), output_string, self.suffix_line_edit.text())


class SearchReplaceWidget(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        super(SearchReplaceWidget, self).__init__(*args, **kwargs)
        main_layout = QtWidgets.QHBoxLayout()
        main_layout.setContentsMargins(0, 0, 0, 0)
        self.search_line = QtWidgets.QLineEdit()
        self.search_line.setPlaceholderText("Search")
        self.replace_line = QtWidgets.QLineEdit()
        self.replace_line.setPlaceholderText("Replace")
        main_layout.addWidget(self.search_line)
        main_layout.addWidget(self.replace_line)

        self.remove_button = QtWidgets.QPushButton("-")
        main_layout.addWidget(self.remove_button)
        self.setLayout(main_layout)



from PyQt6.QtCore import QModelIndex, Qt
from PyQt6.QtGui import QColor, QPalette, QPainter, QBrush
from PyQt6.QtWidgets import QStyledItemDelegate, QStyle, QStyleOptionViewItem


class SelectedHighlighter(QStyledItemDelegate):
    """Delegate for changing highlighting behavior in QTableviews"""

    color_default = QColor("#CFAAFF")

    _test_color = '#FFAAFB'

    _properties = {'foreground': {'selected': {'active': "#000000",
                                               'inactive': "#000000"},
                                  'unselected': {'active': "#000000",
                                                 'inactive': "#000000"}
                                  },
                   'background': {'selected': {'active': "#B8B8B8",
                                               'inactive': "#B8B8B8"},
                                  'unselected': {'active': "#000000",
                                                 'inactive': "#000000"}}}

    def __init__(self, parent=None):
        super().__init__(parent)

    def paint(self, painter: QPainter, option: 'QStyleOptionViewItem', index: QModelIndex) -> None:

        if option.state & QStyle.StateFlag.State_Selected:  # color mixing for selected indexes
            item_color = index.data(
                Qt.ItemDataRole.BackgroundRole)  # retrieve background color of the index, if one exists
            mix_color = self._properties.get('background').get('selected').get('active')

            if item_color is not None:
                mixed_color = SelectedHighlighter.combineColors(item_color, mix_color)
                option.palette.setColor(QPalette.ColorRole.Highlight,
                                        mixed_color)  # background color changes if selected

        QStyledItemDelegate.paint(self, painter, option, index)

    def initStyleOption(self, option: QStyleOptionViewItem, index):
        super(SelectedHighlighter, self).initStyleOption(option, index)

        c_selected = QColor(
            self._properties.get('background').get('selected').get('active'))  # coloring of font in the selection
        color_unselect = self._properties.get('foreground').get('unselected').get(
            'active')  # color of the unselected font if available, otherwise default to black

        c_unselected = index.data(Qt.ItemDataRole.ForegroundRole)
        c_unselected = c_unselected if isinstance(c_unselected, QColor) else QColor(color_unselect)

        cg = (
            QPalette.ColorGroup.Active if option.state & QStyle.StateFlag.State_Enabled else QPalette.ColorGroup.Disabled)

        if option.state & QStyle.StateFlag.State_Selected:
            option.palette.setColor(cg, QPalette.ColorRole.HighlightedText, c_selected)  # text color for selected

            # option.palette.setColor(cg, QPalette.ColorRole.Highlight, QColor(self._test_color)) #background color for selected

        # option.backgroundBrush = QBrush(QColor('white'))#colors background of listview

        option.palette.setBrush(QPalette.ColorRole.Text, c_unselected)  #
        """
        setBrush(self, cg: QPalette.ColorGroup, cr: QPalette.ColorRole, brush: Union[QBrush, Union[QColor, Qt.GlobalColor, int], QGradient])
        setBrush(self, acr: QPalette.ColorRole, abrush: Union[QBrush, Union[QColor, Qt.GlobalColor, int], QGradient])
        """

    def initStyleOptionTest(self, option: QStyleOptionViewItem, index):
        """Testing Method"""

        super(SelectedHighlighter, self).initStyleOption(option, index)

        # use option.state and QStyle.StateFlags for finer control of coloring

        if option.state & QStyle.StateFlag.State_Selected:
            pass

            # Status type as ColorGroup, then target role as ColorRole
            option.palette.setColor(QPalette.ColorGroup.Active, QPalette.ColorRole.HighlightedText, QColor('#FFAAFB'))
            option.palette.setColor(QPalette.ColorGroup.Disabled, QPalette.ColorRole.HighlightedText, QColor('#BEAAFF'))
            option.palette.setColor(QPalette.ColorGroup.Inactive, QPalette.ColorRole.HighlightedText, QColor('#AAC9FF'))

            option.palette.setColor(QPalette.ColorGroup.Active, QPalette.ColorRole.Text, QColor('#FF00FB'))
            option.palette.setColor(QPalette.ColorGroup.Disabled, QPalette.ColorRole.Text, QColor('#6536FF'))
            option.palette.setColor(QPalette.ColorGroup.Inactive, QPalette.ColorRole.Text, QColor('#3680FF'))

            option.palette.setColor(QPalette.ColorGroup.Active, QPalette.ColorRole.Highlight, QColor('#000000'))
            option.palette.setColor(QPalette.ColorGroup.Disabled, QPalette.ColorRole.Highlight, QColor('#000000'))
            option.palette.setColor(QPalette.ColorGroup.Inactive, QPalette.ColorRole.Highlight, QColor('#000000'))

        option.backgroundBrush = QBrush(QColor(100, 200, 100, 200))  # colors entire background

        """
        QPalette.ColorGroup.Disabled
        QPalette.ColorGroup.Active #AKA Normal
        QPalette.ColorGroup.Inactive
        """

    @staticmethod
    def combineColors(color1, color2) -> QColor:
        color1 = QColor(color1) if not isinstance(color1, QColor) else color1
        color2 = QColor(color2) if not isinstance(color2, QColor) else color2
        mixed_color = QColor()

        # average color channels
        mixed_color.setRed(int((color1.red() + color2.red()) / 2))
        mixed_color.setGreen(int((color1.green() + color2.green()) / 2))
        mixed_color.setBlue(int((color1.blue() + color2.blue()) / 2))

        return mixed_color



import collections
import json
import sys
import time

from PyQt6 import QtCore, QtWidgets, QtGui

from power_qt.json_editor import json_editor_system as system, data_tree, batch_name
from power_qt.json_editor import ui_utils

standalone_app = None
if not QtWidgets.QApplication.instance():
    standalone_app = QtWidgets.QApplication(sys.argv)

import sys
import traceback
from collections import namedtuple

from PyQt6.QtCore import pyqtSignal


def excepthook(exc_type, exc_value, exc_tb):
    enriched_tb = _add_missing_frames(exc_tb) if exc_tb else exc_tb
    # Note: sys.__excepthook__(...) would not work here.
    # We need to use print_exception(...):
    traceback.print_exception(exc_type, exc_value, enriched_tb)


def _add_missing_frames(tb):
    result = fake_tb(tb.tb_frame, tb.tb_lasti, tb.tb_lineno, tb.tb_next)
    frame = tb.tb_frame.f_back
    while frame:
        result = fake_tb(frame, frame.f_lasti, frame.f_lineno, result)
        frame = frame.f_back
    return result


fake_tb = namedtuple('fake_tb', ('tb_frame', 'tb_lasti', 'tb_lineno', 'tb_next'))
sys.excepthook = excepthook


class LocalConstants:
    # modify options
    keys = "Keys"
    values = "Values"
    keys_and_values = "Keys & Values"
    batch_modify_options = [keys_and_values, keys, values]


lk = LocalConstants


class JsonEditorWidget(QtWidgets.QWidget):

    def __init__(self, *args, **kwargs):
        super(JsonEditorWidget, self).__init__(*args, **kwargs)

        self.main_layout = QtWidgets.QVBoxLayout()

        self.active_json_indent_level = None
        self.last_font_change_time = 1

        self.path_widget = ui_utils.QtPathWidget(
            settings_name="JsonEditor",
            file_filter="YAML (*.yaml)",
            recent_paths_amount=100,
            only_show_existing_recent_paths=True,
        )

        self.filter_widget = QtWidgets.QLineEdit()
        self.filter_widget.setPlaceholderText("filter")
        self.filter_widget.setClearButtonEnabled(True)
        self.filter_widget.textEdited.connect(self.filter_data)

        self.data_tree_widget = data_tree.DataTreeWidget()
        self.batch_modify_widget = batch_name.BatchNameWidget()
        self.data_tree_widget.data_is_shown.connect(self.data_visibility_state_changed)

        self.helper_overlay = HelperMessageOverlay(self.data_tree_widget.tree_widget)

        self.data_tree_widget.tree_widget.viewport().installEventFilter(self)
        self.setAcceptDrops(True)

        ###########################################################
        # JSON editor specific ui
        self.modify_hierarchy = QtWidgets.QCheckBox("Modify Hierarchy")
        self.modify_hierarchy.setChecked(False)
        self.modify_type_chooser = QtWidgets.QComboBox()
        self.modify_type_chooser.addItems(lk.batch_modify_options)
        self.modify_rename_button = QtWidgets.QPushButton("Rename")
        self.modify_duplicate_button = QtWidgets.QPushButton("Duplicate")

        # connect signals
        self.modify_rename_button.clicked.connect(self.modify_rename)
        self.modify_duplicate_button.clicked.connect(self.modify_duplicate)

        modify_layout = QtWidgets.QHBoxLayout()
        modify_layout.setContentsMargins(0, 0, 0, 0)
        modify_layout.addWidget(self.modify_hierarchy)
        modify_layout.addWidget(self.modify_type_chooser)
        modify_layout.addWidget(self.modify_rename_button)
        modify_layout.addWidget(self.modify_duplicate_button)
        ###########################################################

        # self.main_layout.addWidget(self.path_widget)
        # self.main_layout.addWidget(self.filter_widget)
        self.main_layout.addWidget(self.data_tree_widget)
        # self.main_layout.addWidget(self.batch_modify_widget)
        # self.main_layout.addLayout(modify_layout)
        self.setLayout(self.main_layout)

        self.path_widget.path_changed.connect(self.load_yaml)

    def modify_rename(self):
        modify_keys = self.modify_type_chooser.currentText() in (lk.keys, lk.keys_and_values)
        modify_values = self.modify_type_chooser.currentText() in (lk.values, lk.keys_and_values)

        items_to_modify = self.data_tree_widget.get_selected_items()

        if self.modify_hierarchy.isChecked():
            item_descendants = []
            for item in items_to_modify:
                item_descendants.extend(data_tree.get_all_item_descendants(item))
            items_to_modify.extend(item_descendants)

        for item in items_to_modify:
            if modify_keys:
                item.setText(
                    data_tree.lk.col_key,
                    self.batch_modify_widget.modify_string(item.text(data_tree.lk.col_key))
                )

            if modify_values and not data_tree.item_supports_children(item):
                item.setText(
                    data_tree.lk.col_value,
                    self.batch_modify_widget.modify_string(item.text(data_tree.lk.col_value))
                )

    def modify_duplicate(self):
        modify_keys = self.modify_type_chooser.currentText() in (lk.keys, lk.keys_and_values)
        modify_values = self.modify_type_chooser.currentText() in (lk.values, lk.keys_and_values)

        for item in self.data_tree_widget.get_selected_items():
            item_data = self.data_tree_widget.get_widget_item_values(item)
            item_key = item.text(data_tree.lk.col_key)
            if modify_keys:
                item_key = self.batch_modify_widget.modify_string(item_key)

            parent = item.parent() if item.parent() else self.data_tree_widget.tree_widget.invisibleRootItem()

            new_item = self.data_tree_widget.add_data_to_widget(
                data_key=item_key,
                data_value=item_data,
                parent_item=parent,
                merge=not data_tree.item_supports_children(item),
                key_safety=True
            )

            # insert next to the original
            item_index = parent.indexOfChild(item)
            parent.takeChild(parent.indexOfChild(new_item))
            parent.insertChild(item_index + 1, new_item)

            items_to_modify = [new_item]
            if self.modify_hierarchy.isChecked():
                items_to_modify.extend(data_tree.get_all_item_descendants(new_item))

            for duped_item in items_to_modify:
                if modify_keys and duped_item != new_item:  # key has already been modified, so can skip that on root
                    duped_item.setText(
                        data_tree.lk.col_key,
                        self.batch_modify_widget.modify_string(duped_item.text(data_tree.lk.col_key))
                    )

                if modify_values and not data_tree.item_supports_children(duped_item):
                    duped_item.setText(
                        data_tree.lk.col_value,
                        self.batch_modify_widget.modify_string(duped_item.text(data_tree.lk.col_value))
                    )

            data_tree.fix_list_indices(parent)

    def filter_data(self):
        filter_text = self.filter_widget.text()

        self.data_tree_widget.set_filter(filter_text, search_columns=(data_tree.lk.col_key, data_tree.lk.col_value))

        if filter_text:
            self.data_tree_widget.tree_widget.expandAll()
        else:
            self.data_tree_widget.tree_widget.expandToDepth(self.data_tree_widget.default_expand_depth)

    def data_visibility_state_changed(self, state):
        self.helper_overlay.setVisible(not state)

    ###############################################################################
    # File Handling
    def load_json(self, path):
        json_data = system.load_json(path)
        if json_data is None:
            return

        self.active_json_indent_level = system.get_json_indent_level(path)
        if self.active_json_indent_level is not None:
            print("found indentation in file, setting to: {}".format(self.active_json_indent_level))

        self.data_tree_widget.set_data(json_data)
        print("Loaded Json from: {}".format(path))

    def load_yaml(self, path):
        json_data = system.load_yaml(path)
        if json_data is None:
            return

        self.data_tree_widget.set_data(json_data)
        self.currentPath = path
        print("Loaded Json from: {}".format(path))

    def new_file(self):
        self.path_widget.set_path("")
        self.data_tree_widget.action_clear()
        self.active_json_indent_level = None

    def save_json(self, path=None):
        data_from_ui = self.data_tree_widget.get_data()
        if data_from_ui == "":
            print("No data found in UI, please define a root type before saving")
            return

        ui_path = self.path_widget.path()
        if path is None:
            if ui_path:
                path = ui_path
            else:
                path = self.path_widget.get_dialog_path()

        system.save_yaml(data_from_ui, path)
        print("Saved Json to: {}".format(path))

        # file was newly saved, set the path in the UI
        if not ui_path:
            self.path_widget.set_path(path)

    def save_json_as(self):
        new_path = self.path_widget.get_dialog_path()
        self.save_json(new_path)
        self.path_widget.set_path(new_path)

    def reload(self):
        self.load_json(self.path_widget.path())

    ###############################################################################
    # Overrides

    def resizeEvent(self, event):
        self.helper_overlay.resize(event.size())
        event.accept()

    def eventFilter(self, qobject, event):
        if event.type() == QtCore.QEvent.Type.Wheel:
            modifiers = QtWidgets.QApplication.keyboardModifiers()
            if modifiers == QtCore.Qt.Modifier.ControlModifier:
                if event.delta() > 0:
                    self.increment_tree_font_size(1)
                else:
                    self.increment_tree_font_size(-1)
                return True

        return False  # standard event processing

    def dragEnterEvent(self, e):
        if e.mimeData().hasText():
            e.accept()
        else:
            e.ignore()

    def dropEvent(self, event):
        if event.mimeData().hasUrls():  # if file or link is dropped
            url = event.mimeData().urls()[0]  # get first url
            self.path_widget.set_path(url.toLocalFile())
        else:
            try:
                text = event.mimeData().text()
                json_data = json.loads(text, object_pairs_hook=collections.OrderedDict)
                self.data_tree_widget.set_data(json_data)
            except Exception as e:
                self.status_message("No JSON serializable data could be read from string")

    ###############################################################################
    # Font Size

    def increment_tree_font_size(self, direction):
        if time.time() - self.last_font_change_time < 0.02:
            return

        current_font = self.data_tree_widget.tree_widget.font()
        new_font_size = current_font.pointSize() + direction
        self.set_tree_font_size(new_font_size)

    def set_tree_font_size(self, size=8):
        current_font = self.data_tree_widget.tree_widget.font()
        current_font.setPointSizeF(size)
        self.data_tree_widget.tree_widget.setFont(current_font)
        self.last_font_change_time = time.time()
        self.helper_overlay.font_size = size + 6



class HelperMessageOverlay(QtWidgets.QWidget):

    def __init__(self, parent=None):
        super(HelperMessageOverlay, self).__init__(parent)
        self.setAttribute(QtCore.Qt.WidgetAttribute.WA_TransparentForMouseEvents)

        palette = QtGui.QPalette(self.palette())
        palette.setColor(palette.ColorRole.Base, QtCore.Qt.GlobalColor.transparent)

        self.font_size = 13
        self.empty_json_message = "Drag and drop a JSON file, \n" \
                                  "or drag in JSON readable data, \n" \
                                  "or paste data from clipboard."

        self.setPalette(palette)

    def paintEvent(self, event):
        painter = QtGui.QPainter()
        painter.begin(self)
        painter.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing)
        painter.fillRect(event.rect(), QtGui.QBrush(QtGui.QColor(100, 100, 100, 100)))
        painter.setFont(QtGui.QFont("seqoe", self.font_size))
        painter.drawText(event.rect(), QtCore.Qt.AlignmentFlag.AlignCenter, self.empty_json_message)
        painter.setPen(QtGui.QPen(QtCore.Qt.PenStyle.NoPen))


class JsonEditorWindow(ui_utils.ToolWindow):
    newConfig = pyqtSignal()
    def __init__(self):
        super(JsonEditorWindow, self).__init__()
        self.ui = JsonEditorWidget()
        self.setCentralWidget(self.ui)
        self.setWindowTitle("JSON Editor")

        menu_bar = QtWidgets.QMenuBar()
        file_menu = menu_bar.addMenu("File")
        # file_menu.setTearOffEnabled(True)
        # file_menu.addAction("New", self.ui.new_file, QtGui.QKeySequence("Ctrl+N"))
        # file_menu.addAction("Open", self.ui.path_widget.open_dialog_and_set_path, QtGui.QKeySequence("Ctrl+O"))
        # file_menu.addAction("Save", self.ui.save_json, QtGui.QKeySequence("Ctrl+S"))
        # file_menu.addAction("Save As...", self.ui.save_json_as, QtGui.QKeySequence("Ctrl+Shift+S"))
        # file_menu.addSeparator()
        # file_menu.addAction("Reload from Disk", self.ui.reload, QtGui.QKeySequence("F5"))
        #
        # edit_menu = menu_bar.addMenu("Edit")
        # edit_menu.setTearOffEnabled(True)
        # edit_menu.addAction(
        #     "Cut",
        #     self.ui.data_tree_widget.action_cut_data_to_clipboard,
        #     QtGui.QKeySequence("Ctrl+X"),
        # )
        #
        # edit_menu.addAction(
        #     "Copy",
        #     self.ui.data_tree_widget.action_copy_data_to_clipboard,
        #     QtGui.QKeySequence("Ctrl+C"),
        # )
        #
        # edit_menu.addAction(
        #     "Paste",
        #     self.ui.data_tree_widget.action_paste_data_from_clipboard,
        #     QtGui.QKeySequence("Ctrl+V"),
        # )
        #
        # edit_menu.addAction(
        #     "Duplicate",
        #     self.ui.data_tree_widget.action_duplicate_selected_items,
        #     QtGui.QKeySequence("Ctrl+D"),
        # )
        #
        # edit_menu.addAction(
        #     "Delete",
        #     self.ui.data_tree_widget.delete_selected_items,
        #     QtGui.QKeySequence("DEL"),
        # )
        #
        # edit_menu.addSeparator()
        # edit_menu.addAction(
        #     "Move Up",
        #     self.ui.data_tree_widget.action_move_selected_items_up,
        #     QtGui.QKeySequence("Alt+Up"),
        # )
        #
        # edit_menu.addAction(
        #     "Move Down",
        #     self.ui.data_tree_widget.action_move_selected_items_down,
        #     QtGui.QKeySequence("Alt+Down"),
        # )
        #
        # edit_menu.addAction(
        #     "Sort Alphabetical",
        #     self.ui.data_tree_widget.sort_selected_items,
        # )
        #
        # edit_menu.addAction(
        #     "Select Hierarchy",
        #     self.ui.data_tree_widget.select_hierarchy,
        #     QtGui.QKeySequence("Ctrl+Down"),
        # )
        #
        # display_menu = menu_bar.addMenu("Display")
        # display_menu.setTearOffEnabled(True)
        # display_menu.addAction(
        #     "Reset Font Size",
        #     self.ui.set_tree_font_size,
        # )
        #
        # display_menu.addAction(
        #     "Increase Font Size",
        #     (lambda: self.ui.increment_tree_font_size(1)),
        #     QtGui.QKeySequence("Ctrl++"),
        # )
        #
        # display_menu.addAction(
        #     "Decrease Font Size",
        #     (lambda: self.ui.increment_tree_font_size(-1)),
        #     QtGui.QKeySequence("Ctrl+-"),
        # )

        self.setMenuBar(menu_bar)

    def closeEvent(self, event):
        self.ui.save_json(self.ui.currentPath)

        self.newConfig.emit()
        self.close()


def main(file_path=None, refresh=False):
    win = JsonEditorWindow()
    win.main(refresh=refresh)
    win.resize(800, 800)

    if file_path is not None:
        win.ui.load_yaml(file_path)

    if standalone_app:
        ui_utils.standalone_app_window = win
        sys.exit(standalone_app.exec())

    return win


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from typing import Annotated

from PyQt6.QtCore import QTimer
from PyQt6.QtWidgets import QPushButton, QMenu

__all__ = ['ActiveButton']

from power_qt.other import DefaultPropertiesWidget


@dataclass
class _button_parameters:
    _original_text: str
    _inprogress_text: str
    _cancelled_text: str
    _error_text: str
    _completed_text: str


button_typing = Annotated[str, _button_parameters('original', 'inprogress', 'cancelled', 'error', 'completed')]


class PropertyContextMenu:
    _properties: dict = {}
    __menu_activated = False

    def setupMenu(self):
        self.context_menu = QMenu(self)
        action1 = self.context_menu.addAction('Customize Tabs')
        action1.triggered.connect(self.showEditor)

        self.editor = DefaultPropertiesWidget(self)
        self.editor.setObjectName('Button Editor')
        self.editor.setWindowTitle('Name')
        self.editor.add_pages(self._properties)
        self.__menu_activated = True

    def _set_default_dictionary(self, new_obj):
        self._properties.update(new_obj)

    def contextMenuEvent(self, a0):
        if self.__menu_activated:
            self.context_menu.exec(a0.globalPos())

    def showEditor(self):
        self.editor.show()


class ActiveButton(QPushButton, PropertyContextMenu):
    """Qt PushButton with messaging responses"""

    _properties = {'Button Text': {'original': '',
                                   'inprogress': 'Running',
                                   'cancelled': 'Operation cancelled',
                                   'error': 'Error Encountered!',
                                   'completed': 'Operation Completed'},
                   'Button Coloring': {'background-color': {'original': '#83fff9',
                                                            'inprogress': '#ceff83',
                                                            'cancelled': '#ffe883',
                                                            'error': '#ff0000',
                                                            'completed': '#83ff87'},
                                       'color': {'original': '#000000',
                                                 'inprogress': '#000000',
                                                 'cancelled': '#000000',
                                                 'error': '#000000',
                                                 'completed': '#000000'}
                                       }}

    def assignStylesheet(self, code):
        bkg_color = self._properties['Button Coloring']['background-color'][code]
        fgd_color = self._properties['Button Coloring']['color'][code]

        self.setStyleSheet('QPushButton {background-color:%s; color:%s}' % (bkg_color, fgd_color))

    def __init__(self, parent=None):
        super().__init__(parent)
        self.__timer = QTimer()
        self.__timer.timeout.connect(self.reset)

        self.setupMenu()

    def setText(self, text):
        super().setText(text)
        self._properties['Button Text']['original'] = text
        self.assignStylesheet('original')

    def running(self):
        super().setText(self._properties['Button Text']['inprogress'])
        self.assignStylesheet('inprogress')
        self.setEnabled(False)

    def cancelled(self):
        super().setText(self._properties['Button Text']['cancelled'])
        self.assignStylesheet('cancelled')
        self.__timer.start(5000)
        self.setEnabled(True)

    def errored(self, *args, **kwargs):
        super().setText(self._properties['Button Text']['error'])
        self.assignStylesheet('error')
        self.setEnabled(True)

    def finished(self):
        super().setText(self._properties['Button Text']['completed'])
        self.assignStylesheet('completed')
        self.__timer.start(5000)
        self.setEnabled(True)

    def reset(self):
        super().setText(self._properties['Button Text']['original'])
        self.assignStylesheet('original')
        self.__timer.stop()

    def setTextLabels(self, target: button_typing):
        """Sets """
        pass



import logging

from PyQt6.QtCore import QModelIndex, QSortFilterProxyModel

logger = logging.getLogger('__main__')

from PyQt6.QtCore import Qt

__all__ = ['myCustomQsort']


class myCustomQsort(QSortFilterProxyModel):
    cache = set()

    def __init__(self, parent=None):
        super().__init__(parent)

    def clearCache(self):
        self.cache = set()

    def filterAcceptsRow(self, source_row: int, source_parent: QModelIndex) -> bool:

        value = self.sourceModel().index(source_row, 0, source_parent)

        inCache = value.data(Qt.ItemDataRole.DisplayRole) in self.cache
        matched = self.filterRegularExpression().match(value.data(self.filterRole())).hasMatch()
        if not inCache and matched:
            self.cache.add(str(value.data(Qt.ItemDataRole.DisplayRole)))
            return True
        else:
            return False



import json
import os
import re
import sys
from collections import OrderedDict
from functools import partial

from PyQt6.QtGui import QColor
from PyQt6.QtWidgets import QTreeWidgetItem
from PyQt6.QtCore import Qt
from matplotlib import colors

from power_qt.json_editor import ui_utils
from power_qt.json_editor.ui_utils import QtCore, QtWidgets


class SpecialTreeItem(QTreeWidgetItem):
    def data(self, index, role):
        if role == Qt.ItemDataRole.DisplayRole:
            return super().data(index, role)

        elif role == Qt.ItemDataRole.BackgroundRole:
            data = super().data(index, Qt.ItemDataRole.DisplayRole)

            if re.search(r'^#(?:[0-9a-fA-F]{3}){1,2}$', str(data)):
                return QColor(data)
            else:
                return None

        else:
            return super().data(index, role)

    def setData(self, index, value, role):
        super().setData(index, value, role)

    def ContrastColor(self, hexColor):
        R, G, B = colors.hex2color(hexColor)

        luminance = (0.299 * R + 0.587 * G + 0.114 * B)

        if (luminance > 0.5):
            d = 0  # // bright colors - black font
        else:
            d = 1  # // dark colors - white font

        return colors.rgb2hex((d, d, d))


class LocalConstants:
    col_key = 0
    col_value = 1
    col_type = 2

    add_types = (str, int, float, bool, dict, list)
    default_add_values = {
        str: "STRING",
        dict: {"key": "value"},
        list: ["list_item"],
    }

    root_add_values = {
        dict: {"key": "value"},
        OrderedDict: {"ordered_key": "ordered_value"},
        list: ["list_item"],
        tuple: ["tuple_item"],
    }

    default_key_name = "KEY"

    list_types = (list, tuple)
    dict_types = (dict, OrderedDict)
    list_type_names = (list.__name__, tuple.__name__)
    dict_type_names = (dict.__name__, OrderedDict.__name__)
    supports_children_types = list_types + dict_types
    supports_children_type_names = list_type_names + dict_type_names
    none_type_name = str(type(None).__name__)

    header_names = ("Key", "Value", "Type")


lk = LocalConstants


class DataTreeWidget(QtWidgets.QWidget):
    data_is_shown = QtCore.pyqtSignal(bool)

    def __init__(self, *args, **kwargs):
        super(DataTreeWidget, self).__init__(*args, **kwargs)
        self.default_expand_depth = 2
        self._root_type = None

        self.tree_widget = QtWidgets.QTreeWidget()
        self.tree_widget.setAlternatingRowColors(True)
        self.tree_widget.setSelectionMode(QtWidgets.QTreeWidget.SelectionMode.ExtendedSelection)

        # right click menu
        self.tree_widget.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
        self.tree_widget.customContextMenuRequested.connect(self.build_tree_context_menu)
        self.tree_widget.header().setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
        self.tree_widget.header().customContextMenuRequested.connect(self.build_tree_context_menu)

        self.tree_widget.setHeaderLabels(lk.header_names)
        self.update_header_display()

        self.main_layout = QtWidgets.QVBoxLayout()
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.addWidget(self.tree_widget)
        self.setLayout(self.main_layout)

    #################################################################################################################
    # the two most important functions

    def set_data(self, data):
        self.tree_widget.clear()
        self.data_is_shown.emit(True)

        self._root_type = type(data)
        self.tree_widget.invisibleRootItem().setData(lk.col_type, QtCore.Qt.ItemDataRole.DisplayRole,
                                                     self._root_type.__name__)

        self.add_data_to_widget(data_value=data, parent_item=self.tree_widget.invisibleRootItem(), merge=True)

        self.tree_widget.expandToDepth(self.default_expand_depth)
        self.update_header_display()

    def get_data(self):
        return self.get_widget_item_values(self.tree_widget.invisibleRootItem())

    #################################################################################################################

    def build_tree_context_menu(self):
        action_list = list()

        if self._root_type is None:
            for add_type in lk.supports_children_types:
                action_list.append({
                    "Set Root Type - {}".format(add_type.__name__): partial(self.set_root_type, add_type)
                })
        else:
            action_list.extend([
                {"Cut": self.action_cut_data_to_clipboard},
                {"Copy": self.action_copy_data_to_clipboard},
                {"Paste": self.action_paste_data_from_clipboard},
                {"Duplicate": self.action_duplicate_selected_items},
                {"Delete": self.delete_selected_items},
                "-",
                {"Move Up": self.action_move_selected_items_up},
                {"Move Down": self.action_move_selected_items_down},
                {"Sort Alphabetical": self.sort_selected_items},
                "-",
            ])

            for add_type in lk.add_types:
                action_list.append({
                    "Add - {}".format(add_type.__name__): partial(self.add_item_of_type, add_type)
                })

        ui_utils.build_menu_from_action_list(action_list)

    def set_filter(self, filter_text, search_columns=(lk.col_key,)):
        all_items = self.get_all_items()
        if not filter_text:
            for item in all_items:
                item.setHidden(False)
            return

        # search keys or values
        for item in all_items:
            item.setHidden(True)
        for item in all_items:
            if any(filter_text.lower() in item.text(col).lower() for col in search_columns):
                self.recursive_set_visible(item)

    def recursive_set_visible(self, item):
        item.setHidden(False)
        if item.parent():
            self.recursive_set_visible(item.parent())

    def get_all_items(self, widget=None, widget_list=None):
        if widget is None:
            widget = self.tree_widget.invisibleRootItem()
        if widget_list is None:
            widget_list = list()

        widgets = get_sub_widgets(widget)
        widget_list.extend(widgets)

        for sub_widget in widgets:
            self.get_all_items(sub_widget, widget_list)

        return widget_list

    def action_cut_data_to_clipboard(self):
        self.action_copy_data_to_clipboard()
        self.delete_selected_items()

    def action_copy_data_to_clipboard(self):
        selected_data = self.get_selected_data()
        if not selected_data:
            return

        # Save to clipboard
        str_data = json.dumps(selected_data)
        cb = QtWidgets.QApplication.clipboard()
        cb.setText(str_data, mode=cb.Clipboard)

    def action_paste_data_from_clipboard(self):
        cb = QtWidgets.QApplication.clipboard()
        str_data = cb.text()
        try:
            clipboard_data = json.loads(str_data, object_pairs_hook=OrderedDict)
        except ValueError:
            self.status_message("No JSON serializable data could be read from clipboard")
            return

        if self._root_type == None:
            self.set_data(clipboard_data)
        else:
            self.add_data_to_selected(clipboard_data, merge=True)
        self.data_is_shown.emit(True)

    def action_duplicate_selected_items(self):
        for item in self.get_selected_items():  # type: QtWidgets.QTreeWidgetItem
            item_data = self.get_widget_item_values(item)
            item_key = item.text(lk.col_key)
            parent = self.get_parent(item)

            new_item = self.add_data_to_widget(
                data_key=item_key,
                data_value=item_data,
                parent_item=parent,
                merge=not item_supports_children(item),
                key_safety=True
            )

            # insert next to the original
            item_index = parent.indexOfChild(item)
            parent.takeChild(parent.indexOfChild(new_item))
            parent.insertChild(item_index + 1, new_item)

            fix_list_indices(parent)

    def delete_selected_items(self):
        to_delete = self.get_selected_items()
        for item in to_delete:
            if item is self.tree_widget.invisibleRootItem():
                continue

            item_parent = self.get_parent(item)
            item_parent.removeChild(item)
            fix_list_indices(item_parent)

        if self.tree_widget.invisibleRootItem().childCount() == 0:
            # if we've gotten rid of everything do a full clear
            self.action_clear()

    def sort_selected_items(self):
        selected_items = self.get_selected_items(root_on_empty=False)
        common_parent_map = {}
        for selected_item in selected_items:
            item_parent = self.get_parent(selected_item)

            items_of_parent = common_parent_map.get(item_parent, list())
            items_of_parent.append(selected_item)
            common_parent_map[item_parent] = items_of_parent

        for parent, child_items in common_parent_map.items():
            sorted_children = sorted(child_items, key=lambda x: x.text(lk.col_key))
            first_child_index = min(parent.indexOfChild(child) for child in child_items)
            [parent.takeChild(parent.indexOfChild(child)) for child in child_items]

            sorted_children.reverse()
            [parent.insertChild(first_child_index, sorted_child) for sorted_child in sorted_children]

    def select_hierarchy(self):
        for item in self.get_selected_items():
            item_descendants = get_all_item_descendants(item)
            for item in item_descendants:
                item.setSelected(True)

    def action_clear(self):
        self._root_type = None
        self.tree_widget.clear()
        self.data_is_shown.emit(False)

    def set_root_type(self, add_type):
        root_data = lk.root_add_values.get(add_type, add_type())
        self.set_data(root_data)

    def add_item_of_type(self, add_type=str):
        data_to_add = lk.default_add_values.get(add_type, add_type())
        self.add_data_to_selected(data_to_add)

    def add_data_to_selected(self, data_to_add, merge=False):
        for selected_item in self.get_selected_items():
            if not item_supports_children(selected_item):
                continue

            self.add_data_to_widget(data_value=data_to_add, parent_item=selected_item, merge=merge, key_safety=True)
            fix_list_indices(selected_item)

    def get_selected_items(self, root_on_empty=True):
        selected_items = self.tree_widget.selectedItems()
        if not selected_items and root_on_empty:
            selected_items = [self.tree_widget.invisibleRootItem()]
        return selected_items

    def get_selected_data(self):
        """
        Gets data from all selected items

        :return:
        """
        selected_items = self.get_selected_items()
        if len(selected_items) == 1 and selected_items[0] is self.tree_widget.invisibleRootItem():
            return self.get_data()

        # find most common parent type of all selected items
        parent_data_types = []
        for item in selected_items:
            item_parent = self.get_parent(item)

            parent_data_type = get_data_type(item_parent)
            parent_data_types.append(parent_data_type)

        most_common_parent_type = max(set(parent_data_types), key=parent_data_types.count)

        # build list or dict to put selected data in
        output_data = convert_type_string_to_instance(most_common_parent_type)

        for item in selected_items:
            selected_item_data = self.get_widget_item_values(item)
            if most_common_parent_type in lk.dict_type_names:
                output_data[item.text(lk.col_key)] = selected_item_data
            else:
                output_data.append(selected_item_data)

        return output_data

    def update_header_display(self):
        tree_header = self.tree_widget.header()  # type: QtWidgets.QHeaderView
        tree_header.setStretchLastSection(False)
        tree_header.setSectionResizeMode(1, QtWidgets.QHeaderView.ResizeMode.Stretch)
        self.tree_widget.resizeColumnToContents(lk.col_key)

    def get_widget_item_values(self, widget_item):
        data_type = get_data_type(widget_item)
        data_value = get_data_as_correct_type(widget_item)

        if data_type in lk.dict_type_names:
            for sub_widget in get_sub_widgets(widget_item):
                data_value[sub_widget.text(lk.col_key)] = self.get_widget_item_values(sub_widget)

        elif data_type in lk.list_type_names:
            for sub_widget in get_sub_widgets(widget_item):
                data_value.append(self.get_widget_item_values(sub_widget))

        return data_value

    def getLevel(self, parent_item):
        complete = False
        i = 0

        while not complete:
            if parent_item == self.tree_widget.invisibleRootItem():
                return i

            else:
                i += 1

    def add_data_to_widget(self, data_key="", data_value=None, parent_item=None, merge=False, key_safety=False):

        data_type_name = type(data_value).__name__
        return_item = None

        if key_safety:
            parent_type = get_data_type(parent_item)

            if parent_type in lk.dict_type_names:
                data_key = get_unique_dict_key(parent_item, key_name=data_key)

            if parent_type in lk.list_type_names:
                data_key = "[{}]".format(parent_item.childCount())

        if isinstance(data_value, lk.dict_types):
            if not merge:
                parent_item = QtWidgets.QTreeWidgetItem(
                    parent_item,
                    [data_key, "-------- {} items --------".format(len(data_value)), data_type_name],
                )
                parent_item.setFlags(parent_item.flags())
                return_item = parent_item

            for k, v in data_value.items():
                self.add_data_to_widget(
                    data_key=k,
                    data_value=v,
                    parent_item=parent_item,
                    merge=False,
                    key_safety=key_safety,
                )

        elif isinstance(data_value, lk.list_types):
            if not merge:
                parent_item = QtWidgets.QTreeWidgetItem(
                    parent_item,
                    [data_key, "-------- {} items --------".format(len(data_value)), data_type_name],
                )
                parent_item.setFlags(parent_item.flags() | QtCore.Qt.ItemFlag.ItemIsEditable)
                return_item = parent_item

            for i, v in enumerate(data_value):
                self.add_data_to_widget(
                    data_key="[{}]".format(i),
                    data_value=v,
                    parent_item=parent_item,
                    merge=False,
                    key_safety=key_safety,
                )
        else:

            widget_item = SpecialTreeItem(parent_item, [data_key, str(data_value), data_type_name])

            widget_item.setData(lk.col_value, QtCore.Qt.ItemDataRole.DisplayRole, data_value)
            widget_item.setFlags(widget_item.flags() | QtCore.Qt.ItemFlag.ItemIsEditable)
            return_item = widget_item

        return return_item

    def action_move_selected_items_up(self):
        self.reorder_selected_items(direction=-1)

    def action_move_selected_items_down(self):
        self.reorder_selected_items(direction=1)

    def reorder_selected_items(self, direction=1):
        """
        This thing is a mess

        :param direction:
        :return:
        """
        resolve_parent_list_items = []
        selected_items = self.get_selected_items(root_on_empty=False)
        if direction == 1:
            selected_items.reverse()

        for i, item in enumerate(selected_items):  # Find new indices for selected items
            parent_item = self.get_parent(item)

            was_expanded = item.isExpanded()

            current_index = parent_item.indexOfChild(item)
            new_index = current_index + direction

            if new_index < 0:
                new_index = 0
            if new_index > parent_item.childCount() - 1:
                new_index = parent_item.childCount() - 1

            parent_item.takeChild(current_index)
            parent_item.insertChild(new_index, item)  # ReOrder

            resolve_parent_list_items.append(parent_item)
            item.treeWidget().setCurrentItem(item)  # Set highlight focus
            item.setExpanded(was_expanded)
            parent_item.setExpanded(True)

        if direction == 1:
            selected_items.reverse()  # reverse back

        for item in selected_items:  # ReSelect
            item.setSelected(True)

        for resolved_parent in resolve_parent_list_items:
            fix_list_indices(resolved_parent)

    def get_parent(self, item):
        item_parent = item.parent()
        if item_parent is None:
            item_parent = self.tree_widget.invisibleRootItem()
        return item_parent


def get_sub_widgets(tree_widget_item):
    return [tree_widget_item.child(i) for i in range(tree_widget_item.childCount())]


def get_all_item_descendants(tree_widget_item, item_list=None):
    if item_list is None:
        item_list = []

    item_children = get_sub_widgets(tree_widget_item)

    # This list is recursively extended. So mutability comes into play here.
    item_list.extend(item_children)

    for item_child in item_children:
        get_all_item_descendants(item_child, item_list)

    return item_list


def get_data_as_correct_type(tree_widget_item):
    data_value = tree_widget_item.text(lk.col_value)
    data_type = tree_widget_item.text(lk.col_type)

    if data_type == "NoneType":
        return None

    if data_type in ("int", "float"):
        try:  # This is kinda dumb, but check if the values in the fields are valid
            if data_type == "int":
                int(data_value)
            if data_type == "float":
                float(data_value)
        except ValueError:
            data_value = "".join([s for s in str(data_value) if not s.isalpha()])  # remove letters

    if data_type == "bool":
        # I let you be really sloppy with typing here
        if data_value.lower().startswith("t"):
            return True
        elif data_value.lower() in ["1", "y"]:
            return True
        elif data_value.lower().startswith("f"):
            return False
        else:
            return False

    if data_type in lk.supports_children_type_names:
        eval_statement = "{}()".format(data_type)
    else:
        eval_statement = "{}('{}')".format(data_type, data_value)

    data_of_correct_type = eval(eval_statement)

    return data_of_correct_type


def convert_type_string_to_instance(data_type):
    eval_statement = "{}()".format(data_type)
    data_of_correct_type = eval(eval_statement)
    return data_of_correct_type


def get_data_type(tree_widget_item):
    return tree_widget_item.text(lk.col_type)


def item_supports_children(item):
    return get_data_type(item) in lk.supports_children_type_names


def get_unique_dict_key(parent_item, key_name=""):
    """
    Creates a unique key to use in a dictionary
    +1 to end of name if key already exists

    :param parent_item:
    :param key_name:
    :return:
    """
    if key_name == "":
        key_name = "KEY_0"

    existing_keys = [i.text(lk.col_key) for i in get_sub_widgets(parent_item)]
    if key_name and key_name not in existing_keys:
        return key_name

    if not key_name[-1].isdigit():
        key_name += "_0"

    while key_name in existing_keys:
        old_index = re.findall(r'\d+', key_name)[-1]
        key_name_without_last_number = key_name[:-len(str(old_index))]
        key_name = key_name_without_last_number + str(int(old_index) + 1)

    return key_name


def fix_list_indices(parent_item):
    """
    Sets the list indices on a list tree widget item

    :param parent_item:
    :return:
    """
    if get_data_type(parent_item) in lk.supports_children_type_names:
        parent_item.setText(lk.col_value, "-------- {} items --------".format(parent_item.childCount()))

    if get_data_type(parent_item) in lk.list_type_names:
        children = get_sub_widgets(parent_item)
        for i, child_item in enumerate(children):
            child_item.setData(lk.col_key, QtCore.Qt.ItemDataRole.DisplayRole, "[{}]".format(i))


def test_data_tree():
    app = QtWidgets.QApplication(sys.argv)
    win = QtWidgets.QMainWindow()

    tree = DataTreeWidget()

    win.setCentralWidget(tree)
    win.show()
    win.resize(1000, 1000)

    example_json_path = os.path.join(os.path.dirname(__file__), "resources", "example_json_data.json")
    with open(example_json_path, "r") as fp:
        test_data = json.load(fp, object_pairs_hook=OrderedDict)

    tree.set_data(test_data)
    ui_data = tree.get_data()

    # round trip through the UI, should be identical
    print("TEST_DATA", test_data)
    print("UI_DATA  ", ui_data)
    assert (ui_data == test_data)

    sys.exit(app.exec_())


if __name__ == '__main__':
    test_data_tree()



from PyQt6.QtCore import QUrl
from PyQt6.QtGui import QIcon
from PyQt6.QtWebEngineWidgets import QWebEngineView
from PyQt6.QtWidgets import QMainWindow
from _Utilities import Core, parseTree
from widgets.docs_window import Ui_help_window


class DocumentationView(QMainWindow, Ui_help_window):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setupUi(self)

        home_icon = QIcon(str(Core.assets / 'fugue-icons-3.5.6' / 'icons' / 'home.png'))
        redo_icon = QIcon(str(Core.assets / 'fugue-icons-3.5.6' / 'icons' / 'arrow.png'))
        undo_icon = QIcon(str(Core.assets / 'fugue-icons-3.5.6' / 'icons' / 'arrow-180.png'))
        printer_icon = QIcon(str(Core.assets / 'fugue-icons-3.5.6' / 'icons' / 'printer.png'))

        self.undo.setIcon(undo_icon)
        self.Home.setIcon(home_icon)
        self.redo.setIcon(redo_icon)
        self.compile_HTML.setIcon(printer_icon)

        myDocBrowser = QWebEngineView()
        myDocBrowser.load(QUrl.fromLocalFile(str(Core.docs / 'directory.html')))

        self.main_bar.addWidget(myDocBrowser)

        self.undo.clicked.connect(myDocBrowser.back)
        self.redo.clicked.connect(myDocBrowser.forward)
        self.Home.clicked.connect(lambda: myDocBrowser.load(QUrl.fromLocalFile(str(Core.docs / 'directory.html'))))
        self.compile_HTML.clicked.connect(self.combine_docs)



import pandas as pd
import polars as pl
from PyQt6.QtCore import Qt
from PyQt6.QtGui import QKeySequence
from PyQt6.QtWidgets import QTableView, QAbstractItemView, QApplication

from power_qt.item_delegates.basic_delegates import SelectedHighlighter
from power_qt.pandas_models.basic_table_models import BasicPolarsTableModel, BasicPandasTableModel


class reTableView(QTableView):
    """

    Custom Qt Tableview widget for converting Pandas dataframes into
    GUI viewable tables, uses Pandasmodel_Deux by default. Can be updated using setmodel

    """
    _model = None

    def __init__(self, parent=None):
        QTableView.__init__(self, parent)  # Inherit from QWidget
        self.horizontalHeader().setStyleSheet("QHeaderView::section {background-color:#acacac; font-weight: bold;}")
        self.verticalHeader().setStyleSheet("QHeaderView::section {background-color:#acacac; font-weight: bold;}")

        self.toggle_editing(False)
        self._pandas_model = BasicPandasTableModel()
        self._polars_model = BasicPolarsTableModel()

        self._model = self._pandas_model
        self.__current_model_type = 'Pandas'

        self.setModel(self._model)
        self.setItemDelegate(delegate=SelectedHighlighter(self._model))

    def _check_dataframe(self, dataframe: pl.DataFrame | pd.DataFrame):
        if isinstance(dataframe, pl.DataFrame) and self.__current_model_type != 'Polars':
            self.setModel(self._polars_model)
            self._model = self._polars_model
            self.__current_model_type = 'Polars'

        elif isinstance(dataframe, pd.DataFrame) and self.__current_model_type != 'Pandas':
            self.setModel(self._pandas_model)
            self._model = self._pandas_model
            self.__current_model_type = 'Pandas'

        else:
            pass

    def updateModelData(self, dataframe: pd.DataFrame | pl.DataFrame, *args, **kwargs):

        self._check_dataframe(dataframe)

        self.model().beginResetModel()
        self.model().setNewData(dataframe, *args, **kwargs)
        self.model().endResetModel()

    def update_table(self, dataframe: pd.DataFrame | pl.DataFrame, *args, **kwargs):
        self.updateModelData(dataframe, *args, **kwargs)

    def setModel(self, model) -> None:
        super(reTableView, self).setModel(model)

    def setItemDelegate(self, delegate) -> None:
        super(reTableView, self).setItemDelegate(delegate)

    def toggle_editing(self, _bool: bool, custom_=None):
        "Toggles editing to double-click-to-edit"
        if custom_ is not None:
            self.setEditTriggers(custom_)
        else:
            if _bool:
                self.setEditTriggers(QAbstractItemView.EditTrigger.DoubleClicked)
            else:
                self.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)

    def expanding(self, model=None):
        if model is not None:
            m = model.autoColWidth()
        else:
            m = self.model().autoColWidth()

        for index, width in m:
            self.setColumnWidth(index, width)

    def adjust_columns(self, _all=None, column_indicies=None):
        for index, (width, col_name) in enumerate(zip(self.model()._col_width, self.model()._columns)):
            self.setColumnWidth(index, int(max(int(width), int(len(col_name))) * 5))

    def index(self, row, columns, parent=None, *args, **kwargs):
        return self.model().index(row, columns, parent)

    def keyPressEvent(self, event):
        clipboard = QApplication.clipboard()

        if event.matches(QKeySequence.StandardKey.Copy):  # copy
            selectedRows = [r.row() for r in self.selectionModel().selectedIndexes()]
            selectedColumns = [c.column() for c in self.selectionModel().selectedIndexes()]

            s = ''

            for r in range(min(selectedRows), max(selectedRows) + 1):

                # s += self.table.verticalHeaderItem(c).text() + '\t'
                for c in range(min(selectedColumns), max(selectedColumns) + 1):
                    print(r, c)
                    try:
                        s += str(self.model().index(r, c).data(Qt.ItemDataRole.DisplayRole)) + "\t"
                    except AttributeError:
                        s += "\t"

                s = s[:-1] + "\n"  # eliminate last '\t'

            clipboard.setText(s)
            # QTableView.keyPressEvent(self, event)




from collections.abc import Iterable

import matplotlib
from PyQt6.QtWidgets import QWidget, QTabWidget, QVBoxLayout
from matplotlib.axes import Axes
from matplotlib.backends.backend_qt import NavigationToolbar2QT as NavigationToolbar
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as Canvas
from matplotlib.figure import Figure

matplotlib.use('QTAgg')
__all__ = ['MplCanvas', 'MplTabbar']


class MplCanvas(Canvas):
    """
    Qt compatible matplotlib Canvas class, used as the base canvas
    in all widgets that conduct plotting.
    """
    ax: Axes | list[Axes]
    fig: Figure

    def __init__(self, figsize=(8.5, 11),
                 dpi=100,
                 width_ratios=None,
                 height_ratios=None,
                 nrows=1,
                 ncols=1,
                 *args,
                 **kwargs):
        self.fig = Figure(figsize=figsize,
                          dpi=dpi)

        self.ax = self.fig.subplots(nrows=nrows,
                                    ncols=ncols,
                                    width_ratios=width_ratios,
                                    height_ratios=height_ratios,
                                    *args,
                                    **kwargs)

        Canvas.__init__(self, self.fig)
        # Canvas.setSizePolicy(self, QSizePolicy.Expanding, QSizePolicy.Expanding)
        Canvas.updateGeometry(self)

    def _find_axes(self, axes=None):
        try:
            for ax in axes:
                if not isinstance(ax, Iterable):
                    yield ax
                else:
                    yield from self._find_axes(ax)
        except TypeError:
            yield axes

    def cla(self):
        """Recursively clear all axes on a canvas"""

        for ax in self._find_axes([self.ax]):
            ax.cla()


class MplNavCanvas(QWidget):
    fig: Figure
    ax: Axes | list[Axes]

    def __init__(self,
                 parent=None,
                 figsize=(8.5, 11),
                 dpi=100,
                 width_ratios=None,
                 height_ratios=None,
                 nrows=1,
                 ncols=1,
                 *args,
                 **kwargs):
        super().__init__(parent)

        self.canvas = MplCanvas(figsize=figsize,
                                dpi=dpi,
                                width_ratios=width_ratios,
                                height_ratios=height_ratios,
                                nrows=nrows,
                                ncols=ncols,
                                *args,
                                **kwargs)

        self.navigationbar = NavigationToolbar(self.canvas, self)

        self._layout = QVBoxLayout()
        self._layout.addWidget(self.canvas)
        self._layout.addWidget(self.navigationbar)

    @property
    def fig(self):
        return self.canvas.fig

    @property
    def ax(self):
        return self.canvas.ax


class MplTabbar(QTabWidget):
    """
    Qt compatible tab widget with integrated MplCanvas creation
    """

    figure_tabs: list[MplCanvas] = []

    def __init__(self, parent=None):
        super().__init__(parent)

        self.currentChanged.emit(self.currentIndex())

    def addPlot(self, figsize=(8.5, 11),
                dpi=100,
                width_ratios=None,
                height_ratios=None,
                nrows=1,
                ncols=1,
                tabname='Tab',
                *args,
                **kwargs):
        new_wid = QWidget()
        new_lay = QVBoxLayout()
        new_wid.setLayout(new_lay)
        new_lay.setContentsMargins(0, 0, 0, 0)

        new_canvas = MplCanvas(figsize=figsize,
                               dpi=dpi,
                               width_ratios=width_ratios,
                               height_ratios=height_ratios,
                               nrows=nrows,
                               ncols=ncols,
                               *args,
                               **kwargs)

        new_toolbar = NavigationToolbar(new_canvas, new_wid)
        self.figure_tabs.append(new_canvas)
        new_lay.addWidget(new_canvas)
        new_lay.addWidget(new_toolbar)

        super().addTab(new_wid, tabname)

    def _find_axes(self, axes=None):
        try:
            for ax in axes:
                if not isinstance(ax, Iterable):
                    yield ax
                else:
                    yield from self._find_axes(ax)
        except TypeError:
            yield axes

    def cla(self, tab_no=None):
        """Recursively clear all axes on a canvas"""
        canvas = self.figure_tabs[tab_no]
        if tab_no is None:
            for canvas in self.figure_tabs:
                for ax in self._find_axes([canvas.ax]):
                    ax.cla()

        for ax in self._find_axes([canvas.ax]):
            ax.cla()

    def draw(self, tab_no=None):

        if isinstance(tab_no, int):
            self.figure_tabs[tab_no].draw()

        elif tab_no is None:
            for canvas in self.figure_tabs:
                canvas.draw()
        else:
            pass



import collections
import json
import os

import yaml


def load_json(json_path):
    if not os.path.exists(json_path):
        return

    with open(json_path, "r") as fp:
        json_data = json.load(fp, object_pairs_hook=collections.OrderedDict)
    return json_data

def load_yaml(yaml_path):
    if not os.path.exists(yaml_path):
        return

    with open(yaml_path, "r") as fp:
        json_data = yaml.safe_load(fp)
    return json_data


def get_json_indent_level(json_path):
    with open(json_path, "r") as fp:
        first_couple_of_characters = fp.readline(8)

    if len(first_couple_of_characters) > 2:
        return None

    with open(json_path, "r") as fp:
        fp.readline()  # read first line so we can get to the second one

        next_line = fp.readline()
        indent_number = len(next_line) - len(next_line.lstrip())
    return indent_number


def save_json(json_data, json_path, indent=2):
    with open(json_path, "w+") as fp:
        json.dump(json_data, fp, indent=indent)


def save_yaml(yaml_data, yaml_path):
    with open(yaml_path, "w+") as fp:
        yaml.dump(yaml_data, fp)



from PyQt6.QtWidgets import QWidget, QHBoxLayout, QPushButton, QLineEdit, QFileDialog


class FileBrowser(QWidget):
    start_dir = None
    filter = None

    class modes:
        SingleFile = 0
        MultiFile = 1
        Directory = 2

    def __init__(self, parent=None):
        super().__init__(parent)

        self.layout = QHBoxLayout()

        self.setLayout(self.layout)

        self.browseButton = QPushButton("Browse")
        self.browsedFiles = QLineEdit()

        self.layout.addWidget(self.browseButton)
        self.layout.addWidget(self.browsedFiles)

        self._saveDialog = QFileDialog()

    def savePaths(self):
        return self.browsedFiles.text()

    def setMode(self, mode: modes):
        if mode == self.modes.SingleFile:
            self.browseButton.clicked.connect(self._browser_pointer_file)

        elif mode == self.modes.MultiFile:
            self.browseButton.clicked.connect(self._browser_pointer_multifile)

        elif mode == self.modes.Directory:
            self.browseButton.clicked.connect(self._browser_pointer_folder)

        else:
            self.browseButton.clicked.connect(self._browser_pointer_file)

    def setFileBrowserSettings(self, filter_string=None,
                               start_dir=None):

        self.filter = filter_string
        self.start_dir = start_dir

    def _browser_pointer_file(self):
        filenames, filters = self._saveDialog.getOpenFileName(caption="Select File",
                                                              filter=self.filter,
                                                              directory=self.start_dir)
        self.browsedFiles.setText(filenames)

    def _browser_pointer_multifile(self):
        filenames, filters = self._saveDialog.getOpenFileNames(caption="Select File/s",
                                                               filter=self.filter,
                                                               directory=self.start_dir)
        self.browsedFiles.setText(' | '.join(filenames))

    def _browser_pointer_folder(self):
        filenames = self._saveDialog.getExistingDirectory(caption="Select Folder")
        self.browsedFiles.setText(filenames)



import logging

from PyQt6.QtCore import pyqtSlot,Qt
from PyQt6.QtGui import QFont
from PyQt6.QtWidgets import QWidget, QAbstractItemView, QPushButton, QVBoxLayout, QLabel, QHBoxLayout

logger = logging.getLogger('__main__')


# future version, create ui file for widget

class TwoListSelection(QWidget):
    """
    Custom two list selection widget, facilitates moving of items
    from left to right lists and vice versa

    """

    def __init__(self, parent=None):
        super(TwoListSelection, self).__init__(parent)

    def setup_layout(self, priority=False, myListWidget=None):
        lay = QHBoxLayout(self)
        self.mInput = myListWidget()
        self.mInput.setObjectName('Selected')

        self.mOuput = myListWidget()
        self.mOuput.setObjectName('Available')

        self.mInput.setSelectionMode(QAbstractItemView.SelectionMode(2))
        self.mOuput.setSelectionMode(QAbstractItemView.SelectionMode(2))

        self.mButtonToSelected = QPushButton(">>")
        self.mBtnMoveToAvailable = QPushButton(">")
        self.mBtnMoveToSelected = QPushButton("<")
        self.mButtonToAvailable = QPushButton("<<")

        vlay = QVBoxLayout()
        vlay.addStretch()
        vlay.addWidget(self.mButtonToSelected)
        vlay.addWidget(self.mBtnMoveToAvailable)
        vlay.addWidget(self.mBtnMoveToSelected)
        vlay.addWidget(self.mButtonToAvailable)
        vlay.addStretch()

        self.mBtnUp = QPushButton("Up")
        self.mBtnDown = QPushButton("Down")

        vlay2 = QVBoxLayout()
        vlay2.addStretch()
        vlay2.addWidget(self.mBtnUp)
        vlay2.addWidget(self.mBtnDown)
        vlay2.addStretch()

        vlay_available = QVBoxLayout()
        label_avail = QLabel()
        label_avail.setText('Available Items')
        label_avail.setFont(QFont('Helvetica', 12))

        vlay_available.addWidget(label_avail)
        vlay_available.addWidget(self.mInput)

        vlay_selected = QVBoxLayout()
        self.label_sel = QLabel()
        self.label_sel.setText('Selected Items')
        self.label_sel.setFont(QFont('Helvetica', 12))

        vlay_selected.addWidget(self.label_sel)
        vlay_selected.addWidget(self.mOuput)

        lay.addLayout(vlay_available)
        lay.addLayout(vlay)
        lay.addLayout(vlay_selected)

        if priority:
            lay.addLayout(vlay2)

        self.update_buttons_status()
        self.connections()

    @pyqtSlot()
    def update_buttons_status(self):

        if self.sender() is not None:
            sender_name = self.sender().objectName()
        else:
            sender_name = ''

        self.mBtnUp.setDisabled(not bool(self.mOuput.selectedItems()) or self.mOuput.currentRow() == 0)
        self.mBtnDown.setDisabled(not bool(self.mOuput.selectedItems()) or self.mOuput.currentRow() == (self.mOuput.count() - 1))
        self.mBtnMoveToAvailable.setDisabled(not bool(self.mInput.selectedItems()) or self.mOuput.currentRow() == 0)
        self.mBtnMoveToSelected.setDisabled(not bool(self.mOuput.selectedItems()))

        if sender_name == 'Selected' and self.mOuput.currentItem() is not None:
            self.mOuput.blockSignals(True)
            self.mOuput.selectionModel().clear()
            self.mOuput.blockSignals(False)

        if sender_name == 'Available' and self.mInput.currentItem() is not None:
            self.mInput.blockSignals(True)
            self.mInput.selectionModel().clear()
            self.mInput.blockSignals(False)

    def connections(self):
        self.mInput.itemSelectionChanged.connect(self.update_buttons_status)
        self.mOuput.itemSelectionChanged.connect(self.update_buttons_status)
        self.mBtnMoveToAvailable.clicked.connect(self.on_mBtnMoveToAvailable_clicked)
        self.mBtnMoveToSelected.clicked.connect(self.on_mBtnMoveToSelected_clicked)
        self.mButtonToAvailable.clicked.connect(self.on_mButtonToAvailable_clicked)
        self.mButtonToSelected.clicked.connect(self.on_mButtonToSelected_clicked)
        self.mBtnUp.clicked.connect(self.on_mBtnUp_clicked)
        self.mBtnDown.clicked.connect(self.on_mBtnDown_clicked)

    @pyqtSlot()
    def on_mBtnMoveToAvailable_clicked(self):
        self.mOuput.blockSignals(True)
        self.mInput.blockSignals(True)

        self.mOuput.addItem(self.mInput.takeItem(self.mInput.currentRow()))

        self.mOuput.blockSignals(False)
        self.mInput.blockSignals(False)

    @pyqtSlot()
    def on_mBtnMoveToSelected_clicked(self):
        self.mOuput.blockSignals(True)
        self.mInput.blockSignals(True)
        self.mInput.addItem(self.mOuput.takeItem(self.mOuput.currentRow()))
        self.mOuput.blockSignals(False)
        self.mInput.blockSignals(False)

    @pyqtSlot()
    def on_mButtonToAvailable_clicked(self):
        self.mOuput.blockSignals(True)
        self.mInput.blockSignals(True)
        while self.mOuput.count() > 0:
            self.mInput.addItem(self.mOuput.takeItem(0))
        self.mOuput.blockSignals(False)
        self.mInput.blockSignals(False)

    @pyqtSlot()
    def on_mButtonToSelected_clicked(self):
        self.mOuput.blockSignals(True)
        self.mInput.blockSignals(True)
        while self.mInput.count() > 0:
            self.mOuput.addItem(self.mInput.takeItem(0))
        self.mOuput.blockSignals(False)
        self.mInput.blockSignals(False)

    @pyqtSlot()
    def on_mBtnUp_clicked(self):
        row = self.mOuput.currentRow()
        currentItem = self.mOuput.takeItem(row)
        self.mOuput.insertItem(row - 1, currentItem)
        self.mOuput.setCurrentRow(row - 1)

    @pyqtSlot()
    def on_mBtnDown_clicked(self):
        row = self.mOuput.currentRow()
        currentItem = self.mOuput.takeItem(row)
        self.mOuput.insertItem(row + 1, currentItem)
        self.mOuput.setCurrentRow(row + 1)

    def addAvailableItems(self, items):
        self.mInput.addItems(items)

        self.mInput.setSortingEnabled(True)
        self.mOuput.setSortingEnabled(True)

    def get_left_elements(self):
        r = []
        for i in range(self.mInput.count()):
            it = self.mInput.item(i)
            r.append(it.text())
        return r

    def get_right_elements(self):
        r = []
        for i in range(self.mOuput.count()):
            it = self.mOuput.item(i)
            r.append(it.text())
        return r

    def changeSelectionMode(self, selection_mode: int):
        self.mInput.setSelectionMode(QAbstractItemView.SelectionMode(selection_mode))
        self.mOuput.setSelectionMode(QAbstractItemView.SelectionMode(selection_mode))

    def getCurrentItems(self):
        selected = self.mOuput.selectedItems() + self.mInput.selectedItems()
        return selected

    def getSingleSelected(self, role = Qt.ItemDataRole.DisplayRole):
        selected = self.mOuput.selectedItems() + self.mInput.selectedItems()

        return selected[0].data(role)



import logging
import re

from openpyxl import Workbook
from openpyxl.styles import PatternFill

logger = logging.getLogger('__main__')


class stylized_excel:
    """
    Class to streamline openpxyl stylizing when saving Excel sheets

    """

    origin_point = (0, 0)

    def __init__(self, data,
                 style_func=None,
                 origin_offset=None,
                 sheetnames=None,
                 ignore_values=False,
                 ignore_fill=True):

        """
        data : Pandas Dataframe or list of dataframes
        style_func : callable(data) or list of callable(data) functions corresponding to data list that returns hex color code or ARGB color code from dataframe index values
        origin_offset : positive value tuple pair for offsetting of origin point from 0,0 {col#1, row#1}
                        eg. origin_offset = (2,5) offsets origin to {col#3, row#6}

        sheetname : list of sheetnames corresponding to given data parameter, must be same length as the data parameter

        ignore_values : Do not write dataframe values into cells. Useful when only coloring is desired
        ignore_fill : Skip coloring of cells, defaults to white background

        """

        self._data = [data] if not isinstance(data, list) else data
        self._styling_func = [style_func] if not isinstance(style_func, list) else style_func

        if origin_offset is not None:
            self.origin_point = origin_offset

        if sheetnames is not None:
            self.sheetnames = sheetnames
        else:
            self.sheetnames = [f'Sheet{i + 1}' for i, _ in enumerate(self._data)]

        self._ignore_fill = ignore_fill
        self._ignore_values = ignore_values

    def create_file(self):

        self.workbook = Workbook()

        for sheet in self.sheetnames:
            self.workbook.create_sheet(sheet)

            self.write_to_sheet(sheet)

        return self

    def write_to_sheet(self, sheetname):

        if sheetname in self.workbook.sheetnames:
            target_sheet = self.workbook[sheetname]

            data = self._data[self.sheetnames.index(sheetname)]
            try:
                style_callable = self._styling_func[self.sheetnames.index(sheetname)]
            except IndexError:
                style_callable = self._styling_func[0]

            # write index first
            for r, index_str in enumerate(data.index):
                # offset 1 row to make room for columns
                target_sheet.cell(column=1, row=r + 2, value=index_str)

            for c, col_str in enumerate(data.columns):
                # offset 1 col to make room for rows
                target_sheet.cell(column=c + 2, row=1, value=col_str)

            rows, cols = data.shape

            for row_ind in range(rows):
                for col_ind in range(cols):
                    check_val = data.iloc[row_ind, col_ind]
                    if not self._ignore_values:
                        target_sheet.cell(column=col_ind + 2, row=row_ind + 2, value=check_val)

                    if not self._ignore_fill:
                        target_sheet.cell(column=col_ind + 2, row=row_ind + 2).fill = PatternFill("solid", fgColor=self.verify_color_format(
                            style_callable(check_val)))
        else:
            return

    def save_file(self, savepath):
        self.workbook.save(savepath)

    @staticmethod
    def verify_color_format(color_string):
        if re.search("^#[A-Za-z0-9]{6}$", color_string) is not None:
            color_string = color_string.replace('#', 'ff')  # default to 100% opacity
        return color_string



from PyQt6.QtCore import Qt, pyqtSignal
from PyQt6.QtWidgets import QListWidget, QAbstractItemView, QWidget, QVBoxLayout, QLabel, QListWidgetItem

class ThumbListWidget(QListWidget):
    currentCountChanged = pyqtSignal()
    trigResort = pyqtSignal()

    """List Widget for holding Movable Items"""

    def __init__(self, type, parent=None):
        super(ThumbListWidget, self).__init__(parent)
        self.setDragDropMode(QAbstractItemView.DragDropMode.DragDrop)
        self.setDefaultDropAction(Qt.DropAction.MoveAction)
        self.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        self.setAcceptDrops(True)

        self.maxItems = 9999

    # def addItems(self, items: list):
    #     for i, data in enumerate(sorted(items)):
    #         myQListWidgetItem = QCustomListWidgetItem(self)
    #         # store the data needed to create/re-create the custom widget
    #         myQListWidgetItem.setData(Qt.ItemDataRole.UserRole, (i, data))
    #         self.addItem(myQListWidgetItem)

    # def handleRowsOrdered(self, parent, first, last):
    #     for index in range(first, last + 1):
    #         item = self.item(index)
    #         if item is not None and self.itemWidget(item) is None:
    #             index, name = item.data(Qt.ItemDataRole.UserRole)
    #             widget = QCustomQWidget()
    #             widget.setTextUp(name)
    #             item.setSizeHint(widget.sizeHint())
    #             self.setItemWidget(item, widget)
    #
    #
    # def handleRowsInserted(self, parent, first, last):
    #     for index in range(first, last + 1):
    #         item = self.item(index)
    #
    #         if item is not None and self.itemWidget(item) is None:
    #             index, name = item.data(Qt.ItemDataRole.UserRole)
    #             widget = QCustomQWidget()
    #             widget.setTextUp(name)
    #             item.setSizeHint(widget.sizeHint())
    #             self.setItemWidget(item, widget)
    #             self.currentCountChanged.emit()

    def dropEvent(self, event):
        if self.count() == self.maxItems and event.source() != self:
            event.setDropAction(Qt.IgnoreAction)
        else:
            super().dropEvent(event)

    def getItems(self):
        return [self.item(x).text() for x in range(self.count())]


class QCustomListWidgetItem(QListWidgetItem):
    """Custom List widget item class"""

    def __init__(self, parent=None):
        super(QListWidgetItem, self).__init__(parent)

    def __gt__(self, other):
        return self.data(Qt.ItemDataRole.UserRole)[-1] > other.data(Qt.ItemDataRole.UserRole)[-1]  # or whatever

    def __lt__(self, other):
        return self.data(Qt.ItemDataRole.UserRole)[-1] < other.data(Qt.ItemDataRole.UserRole)[-1]  # or whatever


class QCustomQWidget(QWidget):
    """Custom List widget item class"""

    def __init__(self, parent=None):
        super(QCustomQWidget, self).__init__(parent)
        self.textQVBoxLayout = QVBoxLayout()
        self.textUpQLabel = QLabel()
        self.textQVBoxLayout.addWidget(self.textUpQLabel)
        self.setLayout(self.textQVBoxLayout)

    def setTextUp(self, text):
        self.textUpQLabel.setText(text)

    def gettext(self):
        return self.textUpQLabel.text()



def main(*args, **kwargs):
    from . import json_editor_ui
    return json_editor_ui.main(*args, **kwargs)


def reload_modules():
    import sys
    if sys.version_info.major >= 3:
        from importlib import reload
    else:
        from imp import reload
    
    from . import data_tree
    from . import batch_name
    from . import json_editor_system
    from . import json_editor_ui
    reload(data_tree)
    reload(batch_name)
    reload(json_editor_system)
    reload(json_editor_ui)
    

def startup():
    # from maya import cmds
    # cmds.optionVar(query="") # example of finding a maya optionvar
    pass







import math
from datetime import datetime

import pandas as pd
from PyQt6.QtCore import QDate, QDateTime, QTime
from PyQt6.QtGui import QValidator
from PyQt6.QtWidgets import QDateEdit


class QuarterEdit(QDateEdit):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.steps_new = 3
        self.separator = ' - Q'
        self.setDisplayFormat('yyyy{0}MM')

        self.setDate(datetime.now())

    def dateTimeFromText(self, text: str | None) -> QDateTime:

        year, month = self._stringToDate(text=text)

        add_months = math.ceil((month - 1) * 3)

        date = QDateTime(QDate(year, 1, 1).addMonths(add_months), QTime(self.time()))

        return date

    def textFromDateTime(self, dt: QDateTime | datetime) -> str:
        # derive the quarter from the current date time value and display it

        year = dt.date().year()
        month = dt.date().month()

        quarter = math.ceil(month / 3)

        return f"{year}{self.separator}{quarter:0>1}"

    def stepBy(self, steps: int) -> None:
        if self.currentSection() == QDateEdit.Section.MonthSection:  # custom stepping for the month section to account for quarter
            steps = self.steps_new * steps
            date = self.date()
            self.setDate(date.addMonths(steps))

        else:
            super().stepBy(steps)

    def stepEnabled(self) -> QDateEdit.StepEnabledFlag:
        if self.currentSection() == QDateEdit.Section.MonthSection:
            date = self.date()
            flag = QDateEdit.StepEnabledFlag.StepNone

            if date.month() < 10:  # enable stepping only if Quarter < Q4
                flag = QDateEdit.StepEnabledFlag.StepUpEnabled

            if date.month() > 3:  # enable stepping only if Quarter > Q1
                flag |= QDateEdit.StepEnabledFlag.StepDownEnabled

            return flag

        else:
            return super().stepEnabled()

    def validate(self, input: str | None, pos: int):
        # user input validation
        year_str = False

        year, month = self._stringToDate(text=input)

        new_date = QDate(year, 1, 1).addMonths(month - 1)

        has_separator = input.count(self.separator) == 1
        range_valid = self.minimumDate() < new_date and new_date < self.maximumDate()
        month_in_range = month <= 4

        if has_separator and month_in_range and range_valid:
            state = QValidator.State.Acceptable

        elif has_separator and year_str and len(year_str) < 4:
            state = QValidator.State.Intermediate

        else:
            state = QValidator.State.Intermediate

        return state, input, pos

    def getQuarterLimits(self) -> tuple[datetime, datetime]:
        """Get endpoints of the quarter range, return tuple of the start, end dates"""

        start_date = pd.date_range(end=pd.to_datetime(self.date().toPyDate()), periods=1, freq='QS').date[0]
        end_date = (pd.date_range(start=pd.to_datetime(self.date().toPyDate()), periods=1,
                                  freq='QE') + pd.tseries.offsets.DateOffset(days=1)).date[0]

        return start_date, end_date

    def getYearQuarter(self) -> str:
        """Convert date into string formatted as YYYYQQ"""
        return f'{self.date().year()}Q{self.getQuarter()}'

    def getQuarter(self) -> int:
        """Convert date into integer of quarter"""
        return pd.to_datetime(self.date().toPyDate()).quarter

    def getYQ(self) -> tuple[int, int]:
        """Convert date into tuple of (year,quarter)"""
        return pd.to_datetime(self.date().toPyDate()).year, pd.to_datetime(self.date().toPyDate()).quarter

    def _stringToDate(self, text) -> tuple[int, int]:

        try:
            year_str, month_str = text.split(self.separator)
            month, year = int(month_str), int(year_str)
        except ValueError:
            year, month = 1, 1

        return year, month



from setuptools import find_packages
from setuptools import setup

setup(

    name='Power_Qt',

    author='Dominick Gagliano',

    author_email='dgaglian96@gmail.com',

    version='0.1.3',
    description='A python library for generalized Qt models for PyQt6',
    long_description="Eventually I'll add a description",

    long_description_content_type="text/markdown",
    url='https://github.com/Squidu/Power_Qt',
    install_requires=['superqt>=0.7.6',
                      'pandas~=2.3.2',
                      'numpy~=2.3.2',
                      'sympy~=1.14.0',
                      'polars~=1.33.1',
                      'Pint~=0.25',
                      'matplotlib~=3.10.6',
                      'setuptools~=80.9.0',
                      ],

    packages=find_packages(),

    python_requires='>=3.10,<3.13',

    classifiers=[

        # I can say what phase of development my library is in.
        'Development Status :: 4 - Beta',

        # Here I'll add the audience this library is intended for.
        'Intended Audience :: Developers',

        # Here I'll define the license that guides my library.
        'License :: OSI Approved :: MIT License',

        # Here I'll note that package was written in English.
        'Natural Language :: English',

        # Here I'll note that any operating system can use it.
        'Operating System :: Microsoft :: Windows',

        # Here I'll specify the version of Python it uses.
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',

        # Here are the topics that my library covers.

    ]

)



import pandas as pd
import polars as pl
import polars.selectors as cs
import seaborn
from PyQt6.QtCore import Qt, QAbstractTableModel, QVariant
from PyQt6.QtGui import QColor

__all__ = ['BasicPolarsTableModel', 'BasicPandasTableModel']


class BaseTableModelFunction(QAbstractTableModel):
    _alignment = Qt.AlignmentFlag.AlignLeft
    _styler = None
    _original_data: pd.DataFrame = None
    _style_on_cell = None

    _data = None
    _col_width = None
    _RowCount = 0
    _ColCount = 0
    target_column = 0

    h_index = None
    v_index = None

    palette_list = seaborn.color_palette(None, 20).as_hex()
    palette_dict = dict()

    customRoles = {}

    def getData(self):
        """Returns the private data variable with any edits that were made"""
        return self._data

    def setNewData(self,
                   data,
                   *args,
                   **kwargs):
        """Sets new internal data variable and refreshes model"""

        self._alignment = kwargs.get('align', Qt.AlignmentFlag.AlignLeft)

        self._styler = kwargs.get('styler', None)

        self._style_on_cell = kwargs.get('style_on_cell', False)

        self.target_column = kwargs.get('target_column', 0)

        self.original_data = data

        self.layoutChanged.emit()

    def rowCount(self, parent=None):
        return self._RowCount

    def columnCount(self, parent=None):
        return self._ColCount

    def flags(self, index):
        return Qt.ItemFlag.ItemIsSelectable | Qt.ItemFlag.ItemIsEnabled


class BasicPolarsTableModel(BaseTableModelFunction):
    """
    Custom class to populate a QTableview with a pandas dataframe
    """

    def __init__(self,
                 data=None,
                 styler=None,
                 parent=None,
                 alignment=None,
                 style_on_cell: int = None,
                 *args,
                 **kwargs):

        QAbstractTableModel.__init__(self, parent)

        self.original_data = data
        self._alignment = alignment
        self._styler = styler
        self._style_on_cell = style_on_cell

    @property
    def original_data(self):
        return self._original_data

    @original_data.setter
    def original_data(self, value):
        self._original_data = value if value is not None else pl.DataFrame()
        self._original_data = self._original_data.with_columns((~cs.string()).cast(pl.String))
        self._data = self._original_data

        self._col_width = self._original_data.with_columns(pl.all().str.len_bytes().max().mul(5))

        self._RowCount = self._original_data.height
        self._ColCount = self._original_data.width

        self.v_index = range(self._RowCount)
        self.h_index = self._original_data.columns

    def setData(self, index, value, role=Qt.ItemDataRole.EditRole):
        if index.isValid():
            if role == Qt.ItemDataRole.DisplayRole or role == Qt.ItemDataRole.EditRole:
                self._data[index.row(), index.column()] = value
                self.dataChanged.emit(index, index)
                return True
        return False

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        if index.isValid():
            value = self._data[index.row(), index.column()]

            if self._styler is not None:
                col_lookup = index.column() if self._style_on_cell is None else self._style_on_cell

                color_data = self._data[index.row(), col_lookup]

                if role == Qt.ItemDataRole.BackgroundRole or role == Qt.ItemDataRole.ForegroundRole:
                    return QColor(self._styler(color_data,
                                               role=role,
                                               colname=self.h_index[col_lookup],
                                               fullcol=self._data.row(index.row())))

            if role == Qt.ItemDataRole.DisplayRole:
                return str(value)

            if role == Qt.ItemDataRole.TextAlignmentRole and self._alignment is not None:
                return self._alignment

        return QVariant()

    def headerData(self, index, orientation, role=Qt.ItemDataRole.DisplayRole):
        if role == Qt.ItemDataRole.DisplayRole:
            if orientation == Qt.Orientation.Horizontal:
                return str(self.h_index[index])

            if orientation == Qt.Orientation.Vertical:
                return str(self.v_index[index])

        return QVariant()


class BasicPandasTableModel(BaseTableModelFunction):
    """
    Custom class to populate a QTableview with a pandas dataframe
    """

    def __init__(self,
                 data=None,
                 styler=None,
                 parent=None,
                 alignment=None,
                 style_on_cell: int = None,
                 *args,
                 **kwargs):

        QAbstractTableModel.__init__(self, parent)

        self.original_data = data
        self._alignment = alignment
        self._styler = styler
        self._style_on_cell = style_on_cell

    @property
    def original_data(self):
        return self._original_data

    @original_data.setter
    def original_data(self, value):
        self._original_data = value if value is not None else pd.DataFrame()
        self._data = self._original_data

        self._RowCount, self._ColCount = self._original_data.shape

        self.v_index = self._original_data.index
        self.h_index = self._original_data.columns
        self._col_width = self._original_data.map(lambda x: len(str(x)), na_action='ignore').max(axis=0).to_list()

    def autoColWidth(self):
        widths = []
        for i, col in enumerate(self._data.columns):
            max_char = max(max([len(str(x)) for x in self._data[col].values]), len(col))

            widths.append((i, max_char * 8))

        return widths

    def setData(self, index, value, role=Qt.ItemDataRole.EditRole):
        if index.isValid():
            if role == Qt.ItemDataRole.DisplayRole or role == Qt.ItemDataRole.EditRole:
                self._data.iat[index.row(), index.column()] = value
                self.dataChanged.emit(index, index)
                return True
        return False

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):

        if index.isValid():
            value = self._data.iat[index.row(), index.column()]

            if self._styler is not None:
                col_lookup = index.column() if self._style_on_cell is None else self._style_on_cell

                color_data = self._data.iat[index.row(), col_lookup]

                if role == Qt.ItemDataRole.BackgroundRole or role == Qt.ItemDataRole.ForegroundRole:
                    return QColor(self._styler(color_data,
                                               role=role,
                                               colname=self.h_index[col_lookup],
                                               fullcol=self._data.iloc[index.row()]))

            else:
                if role == Qt.ItemDataRole.BackgroundRole:
                    return QColor('#FFFFFF')

                if role == Qt.ItemDataRole.ForegroundRole:
                    return QColor('#000000')

            if role == Qt.ItemDataRole.DisplayRole:
                return str(value)

            if role == Qt.ItemDataRole.TextAlignmentRole and self._alignment is not None:
                return self._alignment

        return QVariant()

    def headerData(self, index, orientation, role=Qt.ItemDataRole.DisplayRole):
        if role == Qt.ItemDataRole.DisplayRole:
            if orientation == Qt.Orientation.Horizontal:
                return str(self.h_index[index])

            if orientation == Qt.Orientation.Vertical:
                return str(self.v_index[index])

        return QVariant()



import json

from PyQt6.QtCore import pyqtSignal, Qt, QSize, QObject, QEvent
from PyQt6.QtGui import QMouseEvent
from PyQt6.QtWidgets import QTreeView, QVBoxLayout, QTabWidget, QStackedWidget, QDialog, QTabBar, QPushButton, QWidget, \
    QToolButton, QLineEdit

from power_qt.json_editor.data_tree import DataTreeWidget
from power_qt.other.json_models import QJsonTreeItem, QJsonModel


class TabBarPlus(QTabBar):
    """Tab bar that has a plus button floating to the right of the tabs."""

    plusClicked = pyqtSignal()

    def __init__(self,parent=None):
        super().__init__(parent)


        self.lineEditor = QLineEdit()

    def mouseDoubleClickEvent(self, event:QMouseEvent):
        print(event.pos())
        self.lineEditor.setText(str(self.tabText(0)))


# end class MyClass

class CustomTabWidget(QTabWidget):
    """Tab Widget that that can have new tabs easily added to it."""

    def __init__(self, parent=None):
        super().__init__(parent)

        # Tab Bar
        self.tab = TabBarPlus(self)
        self.setTabBar(self.tab)

        self.tabButton = QToolButton(self)
        self.tabButton.setText('+')
        font = self.tabButton.font()
        font.setBold(True)
        self.tabButton.setFont(font)
        self.setCornerWidget(self.tabButton)

    # end Constructor


# end class CustomTabWidget

class QDefaultFieldsModel(QJsonModel):

    def __init__(self, parent=None, edit_columns=None):
        super(QDefaultFieldsModel, self).__init__(parent)

        self.editable_cols = [1]  # edit_columns

        self._rootItem = QJsonTreeItem()
        self._headers = ("key", "value")

    def json(self, root=None):
        """Serialise model as JSON-compliant dictionary

        Arguments:
            root (QJsonTreeItem, optional): Serialise from here
                defaults to the top-level item

        Returns:
            model as dict

        """
        root = root or self._rootItem
        return {self.objectName(): self.genJson(root)}  # modification to return from tabs

    def get_data(self):
        return self.json()

    def genJson(self, item):
        nchild = item.childCount()

        if item.type is dict:
            document = {}
            for i in range(nchild):
                ch = item.child(i)
                document[ch.key] = self.genJson(ch)
            return document

        elif item.type == list:
            document = []
            for i in range(nchild):
                ch = item.child(i)
                document.append(self.genJson(ch))
            return document

        else:
            return item.value


class DefaultFieldsWidget(QDialog):
    attributeUpdated = pyqtSignal(dict)

    def __init__(self, parent=None):
        super().__init__(parent)
        self._isnested = False
        self._models_cache = []
        self._path = None
        self._path_loaded_json = None

        self._lay = QVBoxLayout(self)
        self._tab_block = CustomTabWidget(self)

        self._tab_block.tabButton.clicked.connect(self.add_tab)

        # self._tab_block.tabBar().setTabButton(0, QTabBar.ButtonPosition.RightSide, QPushButton('+'))

        self._lay.addWidget(self._tab_block)

        self.setLayout(self._lay)

        self.resize(600, 1000)
        try:
            self.attributeUpdated.connect(self.save_json)
        except AttributeError:
            pass

        if parent is not None:
            self.setWindowTitle(str(parent))

    def add_tab(self):
        self._createModel(f'new_entry_{self._tab_block.tabBar().count() + 1}', {})

    @property
    def path(self):

        return self._path

    @path.setter
    def path(self, value):
        self._path = value
        try:
            with open(self._path, "r+") as fp:
                self._path_loaded_json = json.load(fp)
        except FileNotFoundError:
            self._path_loaded_json = dict()

    def add_pages(self, new_col_dict):

        if self._path_loaded_json:
            trg_dict = self._path_loaded_json
        else:
            trg_dict = new_col_dict

        self.__clearModels()

        self._isnested = self.isNested(trg_dict)

        if self._isnested:
            for tabname, dictionary in trg_dict.items():
                self._createModel(tabname, dictionary)
        else:
            self._createModel('Dictionary', trg_dict)

    def _createModel(self, name: str, _adict: dict):
        _model = QDefaultFieldsModel()
        _model.load(_adict)
        _model.setObjectName(name)

        _view = QTreeView(self._tab_block)
        _view.setModel(_model)
        _view.setObjectName(name)
        _view.expandAll()

        self._models_cache.append(_model)
        # _view.expandRecursively()
        _view.resizeColumnToContents(0)
        _view.resizeColumnToContents(1)
        _model.dataChanged.connect(self.updatedDictionary)

        self._tab_block.addTab(_view, name)

    def __clearModels(self):
        for tab in range(self._tab_block.count()):
            self._tab_block.removeTab(tab)

        for widget in self._tab_block.findChildren(QTreeView):
            widget.deleteLater()

    def updatedDictionary(self):
        if self._isnested:
            result = {}
            for model in self._models_cache:
                result.update(model.get_data())

        else:
            result = self._models_cache[0].get_data()

        self.attributeUpdated.emit(result)

    def getmyDict(self):
        trees = self._tab_block.findChildren(QStackedWidget)[0].findChildren(QTreeView)
        self._tab_block.objectName()
        d = {}

        for tree in trees:
            d[tree.objectName()] = tree.model().json()

        return d

    @staticmethod
    def isNested(_myDict: dict):
        return any(isinstance(i, dict) for i in _myDict.values())

    def save_json(self, data_from_ui):
        try:
            with open(self._path, "w+") as fp:
                json.dump(data_from_ui, fp)
            print("Saved Json to: {}".format(self._path))
        except:
            pass


class DefaultPropertiesWidget(DefaultFieldsWidget):

    def __init__(self, parent=None):
        super().__init__(parent)
        self.default = dict()

    def setModelDefaults(self, _adict: dict):
        self.default = _adict

    def return_default(self):
        pass

    def setModelSource(self, _dict: dict):
        for key, value in _dict.items():
            self._createModel(key, value)

    def _createModel(self, name: str, _adict: dict):
        _view = DataTreeWidget(self._tab_block)
        _view.set_data(_adict)
        _view.setObjectName(name)

        _view.tree_widget.model().dataChanged.connect(self.updatedDictionary)
        _view.tree_widget.sortByColumn(0, Qt.SortOrder.AscendingOrder)
        self._models_cache.append((name, _view))

        self._tab_block.addTab(_view, name)

    def updatedDictionary(self):
        if self._isnested:
            result = {}
            for name, model in self._models_cache:
                result.update({name: model.get_data()})

        else:
            result = self._models_cache[0].get_data()

        self.attributeUpdated.emit(result)



import sys

from PyQt6.QtCore import QSortFilterProxyModel
from PyQt6.QtWidgets import QWidget, QApplication, QListWidget, QVBoxLayout, QHBoxLayout, QGridLayout

from power_qt.pandas_models.basic_table_models import BasicPandasTableModel
import pandas as pd

from power_qt.tableviews import reTableView


class FilterModel(QWidget):
    _proxy_Base = None
    _dynamic_filters = []

    def __init__(self, parent=None):
        super().__init__(parent)
        self.search_website = QListWidget()
        self.search_priority = QListWidget()
        self.search_name = QListWidget()

        self.createProxies(BasicPandasTableModel(),
                           [(0, self.search_website),
                            (2, self.search_priority),
                            (1, self.search_name)])

    def createProxies(self, model, filtering_widgets):
        """Filtering Widgets as a list of tuples containing (int, widget) pairs"""
        if len(filtering_widgets) > 25:
            raise AttributeError

        self._proxy_Base = model

        self._dynamic_filters.clear()

        for i, (filterColumn, connected_widget) in enumerate(filtering_widgets):
            proxy_name = f'_proxy_{chr(65 + i)}'
            print(proxy_name)

            setattr(self, proxy_name, QSortFilterProxyModel())
            proxy = getattr(self, proxy_name)

            self._dynamic_filters.append((proxy_name, connected_widget))

            proxy.setFilterKeyColumn(filterColumn)  # Search all columns.
            proxy.setSourceModel(getattr(self, '_proxy_Base'))

        self.FinalModel = getattr(self, self._dynamic_filters[-1][0])

    def apply_filter(self):
        for proxy, field in self._dynamic_filters:
            if not field.selectedItems():
                proxy.setFilterFixedString('')  # clearing filter if none is selected
            else:
                _search = '|'.join([item.text() for item in field.selectedItems()])
                proxy.setFilterRegularExpression(_search)  # applying filter


class testFilteringTable(QWidget):

    def __init__(self, parent=None):
        super().__init__(parent)
        self._table = reTableView()

        self.main_layout = QHBoxLayout()
        self._filters_layout = QGridLayout()

        # self.main_layout.addLayout(self._filters_layout)
        self.main_layout.addWidget(self._table)

        self.setLayout(self.main_layout)

        self._filters = []

    def setTable(self, table,):
        self._table.update_table(table)


if __name__ == '__main__':
    df = pd.DataFrame({'num_legs': [2, 4, 8, 0],
                       'num_wings': [2, 0, 0, 0],
                       'num_specimen_seen': [10, 2, 1, 8]},
                      index=['falcon', 'dog', 'spider', 'fish'])

    app_main = QApplication(sys.argv)  # load main window
    win = testFilteringTable()
    win.setWindowTitle('Test Filtering Table')
    win.setTable(df)
    win.show()

    try:
        sys.exit(app_main.exec())
    except SystemExit:
        print('Closing Window...')
